#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""

XmlRpc-on-HttpService-based Export and Import Distribution Providers

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging

# XML RPC modules for Python 2 or 3
try:
    # Python 3
    # pylint: disable=F0401
    from xmlrpc.server import SimpleXMLRPCDispatcher
    import xmlrpc.client as xmlrpclib
    from concurrent.futures import Executor
    from concurrent.futures.thread import ThreadPoolExecutor
except ImportError:
    # Python 2
    # pylint: disable=F0401
    from SimpleXMLRPCServer import SimpleXMLRPCDispatcher
    import xmlrpclib

    Executor = None

from pelix.http import HTTP_SERVICE
from pelix.utilities import to_str

from pelix.ipopo.constants import ARG_BUNDLE_CONTEXT, ARG_PROPERTIES
from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Instantiate,
    Property,
    Requires,
    Invalidate,
    ValidateComponent,
)

from pelix.rsa import prop_dot_suffix
from pelix.rsa.providers.distribution import (
    ExportContainer,
    ImportContainer,
    ExportDistributionProvider,
    ImportDistributionProvider,
    SERVICE_EXPORT_CONTAINER,
    SERVICE_EXPORT_DISTRIBUTION_PROVIDER,
    SERVICE_IMPORT_CONTAINER,
    SERVICE_IMPORT_DISTRIBUTION_PROVIDER,
)

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------
# XmlRpc Distribution Provider Constants. Note that to get interoperability with
# Java-based ECF RSA providers, these must match the Java-side constants.
ECF_XMLRPC_SERVER_CONFIG = "ecf.xmlrpc.server"
ECF_XMLRPC_CLIENT_CONFIG = "ecf.xmlrpc.client"
ECF_XMLRPC_SUPPORTED_CONFIGS = [ECF_XMLRPC_SERVER_CONFIG]
ECF_XMLRPC_NAMESPACE = "ecf.namespace.xmlrpc"
ECF_XMLRPC_SUPPORTED_INTENTS = ["osgi.basic", "osgi.async"]
ECF_XMLRPC_PATH_PROP = "path"
ECF_XMLRPC_TIMEOUT_PROP = "timeout"
ECF_XMLRPC_HOSTNAME_PROP = "hostname"

# ------------------------------------------------------------------------------


class ServerDispatcher(SimpleXMLRPCDispatcher, object):
    """ServerDispatcher (subclass of SimpleXMLRPCDispatcher)
    uses ECF remote service id to identify the service
    for method invocation requests.  See do_POST and _dispatch
    for the actual method invocation.
    """

    def __init__(self, dispatch_func, timeout=None, executor=None):
        super(ServerDispatcher, self).__init__(allow_none=True)
        self._dispatch_func = dispatch_func
        self._timeout = timeout
        self._executor = executor

    def do_POST(self, request, response):
        # pylint: disable=C0103
        data = to_str(request.read_data())
        result = self._marshaled_dispatch(data, self._dispatch)
        response.send_content(200, result, "text/xml")

    def _dispatch(self, method, params):
        obj_method_list = method.split(".")
        if not len(obj_method_list) == 2:
            raise Exception(
                "_dispatch: invalid method="
                + method
                + ".  Must be of form <objectid>.<methodname>"
            )
        # and call _dispatch_func/3
        if self._executor:
            return self._executor.submit(
                self._dispatch_func,
                obj_method_list[0],
                obj_method_list[1],
                params,
            ).result(self._timeout)

        return self._dispatch_func(
            obj_method_list[0], obj_method_list[1], params
        )


# ------------------------------------------------------------------------------
# Implementation of SERVICE_EXPORT_CONTAINER
@ComponentFactory(ECF_XMLRPC_SERVER_CONFIG)
@Provides(SERVICE_EXPORT_CONTAINER)
class XmlRpcExportContainer(ExportContainer):
    """
    Subclass of ExportContainer created by the XmlRpcExportDistributionProvider
    at export time. The @ComponentFactory annotation on this class uses the
    ECF_XMLRPC_SERVER_CONFIG = 'ecf.xmlrpc.server' as it's factory identifier.
    This factory name be the same as the @Property('_config_name',
    'config_name', ECF_XMLRPC_SERVER_CONFIG) in the
    XmlRpcExportDistributionProvider below
    """

    @ValidateComponent(ARG_BUNDLE_CONTEXT, ARG_PROPERTIES)
    def _validate_component(self, bundle_context, container_props):
        # pylint: disable=W0212
        ExportContainer._validate_component(
            self, bundle_context, container_props
        )
        timeout = container_props.get(ECF_XMLRPC_TIMEOUT_PROP, None)
        dp = self._get_distribution_provider()
        if Executor:
            executor = ThreadPoolExecutor(max_workers=5)
        else:
            executor = None
        # register the ServerDispatcher instance our uri_path
        dp._httpservice.register_servlet(
            dp._uri_path,
            ServerDispatcher(self._dispatch_exported, timeout, executor),
        )

    @Invalidate
    def _invalidate_component(self, bundle_context):
        # pylint: disable=W0212
        """
        First invalidate by unregistering the servlet/dispatcher,
        and then call super._invalidate
        """
        try:
            dp = self._get_distribution_provider()
            dp._httpservice.unregister(dp._uri_path)
            ExportContainer._invalidate_component(self, bundle_context)
        except:
            pass


# ------------------------------------------------------------------------------
# Implementation of SERVICE_EXPORT_DISTRIBUTION_PROVIDER
@ComponentFactory("xmlrpc-export-distribution-provider-factory")
@Provides(SERVICE_EXPORT_DISTRIBUTION_PROVIDER)
@Property("_config_name", "config_name", ECF_XMLRPC_SERVER_CONFIG)
@Property("_namespace", "namespace", ECF_XMLRPC_NAMESPACE)
@Property(
    "_supported_configs", "supported_configs", ECF_XMLRPC_SUPPORTED_CONFIGS
)
@Property(
    "_supported_intents", "supported_intents", ECF_XMLRPC_SUPPORTED_INTENTS
)
@Requires("_httpservice", HTTP_SERVICE)
@Property(
    "_uri_path",
    prop_dot_suffix(ECF_XMLRPC_SERVER_CONFIG, ECF_XMLRPC_PATH_PROP),
    "/xml-rpc",
)
@Property(
    "_timeout",
    prop_dot_suffix(ECF_XMLRPC_SERVER_CONFIG, ECF_XMLRPC_TIMEOUT_PROP),
    30,
)
@Property(
    "_hostname",
    prop_dot_suffix(ECF_XMLRPC_SERVER_CONFIG, ECF_XMLRPC_HOSTNAME_PROP),
    "localhost",
)
@Instantiate("xmlrpc-export-distribution-provider")
class XmlRpcExportDistributionProvider(ExportDistributionProvider):
    """
    ExportDistributionProvider subclass.  The _config_name property
    must be set to the same value as the ExportContainer class factory,
    as it is with the XmlRpcExportContainer class above (ECF_XMLRPC_SERVER_CONFIG).
    This is necessary so that at export time, this provider will used
    the factory with name ECF_XMLRPC_SERVER_CONFIG to create ExportContainer
    instances as needed.

    Note that this distribution provider uses the injected _httpservice
    to register servlets via the ExportContainer._validate_component call
    as implemented in XmlRpcExportContainer._validate_component.
    """

    def __init__(self):
        super(XmlRpcExportDistributionProvider, self).__init__()
        self._httpservice = None
        self._uri_path = None
        self._timeout = None
        self._hostname = None

    def _prepare_container_props(self, service_intents, export_props):
        container_props = ExportDistributionProvider._prepare_container_props(
            self, service_intents, export_props
        )
        if self._timeout:
            container_props[ECF_XMLRPC_TIMEOUT_PROP] = self._timeout
        return container_props

    def _prepare_container_id(self, container_props):
        """
        This method is called prior to actual container creation in order to
        create the name/id of the ExportContainer to be subsequently created
        via ipopo.instantiate(ECF_XMLRPC_SERVER_CONFIG,container_id,props).

        The String returned from this method is used in the instantiate call as
        the container_id.
        """
        uri = "http://"
        if self._httpservice.is_https():
            uri = "https://"

        hostname = container_props.get(ECF_XMLRPC_HOSTNAME_PROP, None)
        if not hostname:
            hostname = self._hostname
            if not hostname:
                self._httpservice.get_hostname()

        port = container_props.get("port")
        if not port:
            port = str(self._httpservice.get_access()[1])

        uri = uri + "{0}:{1}".format(hostname, port)
        return uri + self._uri_path


# ------------------------------------------------------------------------------
# Implementation of SERVICE_IMPORT_CONTAINER


@ComponentFactory(ECF_XMLRPC_CLIENT_CONFIG)
@Provides(SERVICE_IMPORT_CONTAINER)
class XmlRpcImportContainer(ImportContainer):
    """
    ImportContainer created via the XmlRpcImportDistributionProvider
    as needed at runtime.  The @ComponentFactory(ECF_XMLRPC_CLIENT_CONFIG) is
    used at to create an instance of this container at import-time.
    """

    def _prepare_proxy(self, endpoint_description):
        """
        This method is called as part of RSA.import_service.  In this case
        an instance of XmlRpcProxy declared below is returned as the object
        representing the imported remote service (proxy).  Once returned,
        this proxy is registered with locals service registry.
        When a consumer retrieves this service and calls a method, the
         __getattr__ method (below) is and this make the remote call
        with a string: <objectid>.<method> using the
        xmlrpc.client.ServerProxy
        """

        class XmlRpcProxy(object):
            # pylint: disable=R0903
            def __init__(self, get_remoteservice_id):
                self._url = get_remoteservice_id[0][1]
                self._rsid = str(get_remoteservice_id[1])

            def __getattr__(self, name):
                return getattr(
                    xmlrpclib.ServerProxy(self._url, allow_none=True),
                    "{0}.{1}".format(self._rsid, name),
                )

        # create instance of XmlRpcProxy and pass in remoteservice id:
        # ((ns,cid),get_remoteservice_id)
        return XmlRpcProxy(endpoint_description.get_remoteservice_id())


# ------------------------------------------------------------------------------
# Implementation of SERVICE_IMPORT_DISTRIBUTION_PROVIDER
@ComponentFactory("xmlrpc-import-distribution-provider-factory")
@Provides(SERVICE_IMPORT_DISTRIBUTION_PROVIDER)
@Property("_config_name", "config_name", ECF_XMLRPC_CLIENT_CONFIG)
@Property("_namespace", "namespace", ECF_XMLRPC_NAMESPACE)
@Property(
    "_supported_configs", "supported_configs", ECF_XMLRPC_SUPPORTED_CONFIGS
)
@Property(
    "_supported_intents", "supported_intents", ECF_XMLRPC_SUPPORTED_INTENTS
)
@Instantiate("xmlrpc-import-distribution-provider")
class XmlRpcImportDistributionProvider(ImportDistributionProvider):
    """
    We get all necessary methods from ImportDistributionProvider
    """

    pass



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
Defines the shell completion handlers for iPOPO concepts

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2
:status: Alpha

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import absolute_import

# Try to import readline
try:
    import readline
except ImportError:
    readline = None

# Add some typing
try:
    # pylint: disable=W0611
    from typing import List
    from pelix.framework import BundleContext
    from pelix.shell.beans import ShellSession
    from .decorators import CompletionInfo
except ImportError:
    pass

# Pelix
from pelix.constants import BundleActivator
from pelix.ipopo.constants import use_ipopo

# Completion classes
from .decorators import (
    SVC_COMPLETER,
    PROP_COMPLETER_ID,
    FACTORY,
    COMPONENT,
    FACTORY_PROPERTY,
)
from .core import Completer

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class ComponentFactoryCompleter(Completer):
    """
    Completes an iPOPO Component factory name
    """

    @staticmethod
    def display_hook(prompt, session, context, matches, longest_match_len):
        # type: (str, ShellSession, BundleContext, List[str], int) -> None
        """
        Displays the available services matches and the service details

        :param prompt: Shell prompt string
        :param session: Current shell session (for display)
        :param context: BundleContext of the shell
        :param matches: List of words matching the substitution
        :param longest_match_len: Length of the largest match
        """
        # Prepare a line pattern for each match (-1 for the trailing space)
        match_pattern = "{{0: <{}}} from {{1}}".format(longest_match_len - 1)

        # Sort matching names
        matches = sorted(match for match in matches)

        # Print the match and the associated name
        session.write_line()
        with use_ipopo(context) as ipopo:
            for factory_name in matches:
                # Remove the spaces added for the completion
                factory_name = factory_name.strip()
                bnd = ipopo.get_factory_bundle(factory_name)
                session.write_line(
                    match_pattern, factory_name, bnd.get_symbolic_name()
                )

        # Print the prompt, then current line
        session.write(prompt)
        session.write_line_no_feed(readline.get_line_buffer())
        readline.redisplay()

    def complete(
        self, config, prompt, session, context, current_arguments, current
    ):
        # type: (CompletionInfo, str, ShellSession, BundleContext, List[str], str) -> List[str]
        """
        Returns the list of services IDs matching the current state

        :param config: Configuration of the current completion
        :param prompt: Shell prompt (for re-display)
        :param session: Shell session (to display in shell)
        :param context: Bundle context of the Shell bundle
        :param current_arguments: Current arguments (without the command itself)
        :param current: Current word
        :return: A list of matches
        """
        # Register a method to display helpful completion
        self.set_display_hook(self.display_hook, prompt, session, context)

        # Return a list of component factories
        with use_ipopo(context) as ipopo:
            return [
                "{} ".format(factory)
                for factory in ipopo.get_factories()
                if factory.startswith(current)
            ]


class ComponentInstanceCompleter(Completer):
    """
    Completes an iPOPO Component instance name
    """

    @staticmethod
    def display_hook(prompt, session, context, matches, longest_match_len):
        # type: (str, ShellSession, BundleContext, List[str], int) -> None
        """
        Displays the available services matches and the service details

        :param prompt: Shell prompt string
        :param session: Current shell session (for display)
        :param context: BundleContext of the shell
        :param matches: List of words matching the substitution
        :param longest_match_len: Length of the largest match
        """
        # Prepare a line pattern for each match (-1 for the trailing space)
        match_pattern = "{{0: <{}}} from {{1}}".format(longest_match_len - 1)

        # Sort matching names
        matches = sorted(match for match in matches)

        # Print the match and the associated name
        session.write_line()
        with use_ipopo(context) as ipopo:
            for name in matches:
                # Remove the spaces added for the completion
                name = name.strip()
                details = ipopo.get_instance_details(name)
                description = "of {factory} ({state})".format(**details)
                session.write_line(match_pattern, name, description)

        # Print the prompt, then current line
        session.write(prompt)
        session.write_line_no_feed(readline.get_line_buffer())
        readline.redisplay()

    def complete(
        self, config, prompt, session, context, current_arguments, current
    ):
        # type: (CompletionInfo, str, ShellSession, BundleContext, List[str], str) -> List[str]
        """
        Returns the list of services IDs matching the current state

        :param config: Configuration of the current completion
        :param prompt: Shell prompt (for re-display)
        :param session: Shell session (to display in shell)
        :param context: Bundle context of the Shell bundle
        :param current_arguments: Current arguments (without the command itself)
        :param current: Current word
        :return: A list of matches
        """
        # Register a method to display helpful completion
        self.set_display_hook(self.display_hook, prompt, session, context)

        # Return a list of component factories
        with use_ipopo(context) as ipopo:
            return [
                "{} ".format(name)
                for name, _, _ in ipopo.get_instances()
                if name.startswith(current)
            ]


class ComponentFactoryPropertiesCompleter(Completer):
    """
    Completes the property names of iPOPO Component factories
    """

    def complete(
        self, config, prompt, session, context, current_arguments, current
    ):
        # type: (CompletionInfo, str, ShellSession, BundleContext, List[str], str) -> List[str]
        """
        Returns the list of services IDs matching the current state

        :param config: Configuration of the current completion
        :param prompt: Shell prompt (for re-display)
        :param session: Shell session (to display in shell)
        :param context: Bundle context of the Shell bundle
        :param current_arguments: Current arguments (without the command itself)
        :param current: Current word
        :return: A list of matches
        """
        with use_ipopo(context) as ipopo:
            try:
                # Find the factory name
                for idx, completer_id in enumerate(config.completers):
                    if completer_id == FACTORY:
                        factory_name = current_arguments[idx]
                        break
                else:
                    # No factory completer found in signature
                    for idx, completer_id in enumerate(config.completers):
                        if completer_id == COMPONENT:
                            name = current_arguments[idx]
                            details = ipopo.get_instance_details(name)
                            factory_name = details["factory"]
                            break
                    else:
                        # No factory name can be found
                        return []

                # Get the details about this factory
                details = ipopo.get_factory_details(factory_name)
                properties = details["properties"]
            except (IndexError, ValueError):
                # No/unknown factory name
                return []
            else:
                return [
                    "{}=".format(key)
                    for key in properties
                    if key.startswith(current)
                ]


# ------------------------------------------------------------------------------


# All completers for this bundle
COMPLETERS = {
    FACTORY: ComponentFactoryCompleter,
    FACTORY_PROPERTY: ComponentFactoryPropertiesCompleter,
    COMPONENT: ComponentInstanceCompleter,
}


@BundleActivator
class _Activator:
    """
    Bundle activator
    """

    def __init__(self):
        self._registrations = []

    def start(self, context):
        # type: (BundleContext) -> None
        """
        Bundle starting

        :param context: The bundle context
        """
        # Register all completers we know
        self._registrations = [
            context.register_service(
                SVC_COMPLETER,
                completer_class(),
                {PROP_COMPLETER_ID: completer_id},
            )
            for completer_id, completer_class in COMPLETERS.items()
        ]

    def stop(self, _):
        # type: (BundleContext) -> None
        """
        Bundle stopping

        :param context: The bundle context
        """
        # Clean up
        for svc_reg in self._registrations:
            svc_reg.unregister()
        del self._registrations[:]



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""

BasicTopologyManager implements TopologyManager API

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging

try:
    # pylint: disable=W0611
    from typing import Any, Dict
    from pelix.framework import ServiceEvent
    from pelix.rsa import RemoteServiceAdminEvent
except ImportError:
    pass

from pelix.ipopo.decorators import ComponentFactory, Instantiate

from pelix.rsa import ECF_ENDPOINT_CONTAINERID_NAMESPACE
from pelix.rsa.providers.discovery import EndpointEvent
from pelix.rsa.topologymanagers import TopologyManager

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


@ComponentFactory("basic-topology-manager-factory")
# Tell iPOPO to instantiate a component instance as soon as the file is loaded
@Instantiate(
    "basic-topology-manager",
    {
        TopologyManager.ENDPOINT_LISTENER_SCOPE: "({0}=*)".format(
            ECF_ENDPOINT_CONTAINERID_NAMESPACE
        )
    },
)
class BasicTopologyManager(TopologyManager):
    """
    BasicTopologyManager extends TopologyManager api
    """

    def event(self, service_event, listener_dict):
        # type: (ServiceEvent, Dict[Any, Any]) -> None
        """
        Implementation of EventListenerHook.  Called by local
        service registry when a service is registered, unregistered
        or modified.  Will be called by thread doing registration/unregister
        service
        """
        self._handle_event(service_event)

    def endpoint_changed(self, endpoint_event, matched_filter):
        # type: (EndpointEvent, Any) -> None
        """
        Implementation of discovery API EndpointEventListener.
        Called by discovery provider when an endpoint change
        ADDED,REMOVED,MODIFIED is detected.  May be called
        by arbitrary thread.
        """
        event_type = endpoint_event.get_type()
        ed = endpoint_event.get_endpoint_description()
        ed_id = ed.get_id()

        if event_type == EndpointEvent.ADDED:
            # if it's an add event, we call handle_endpoint_added
            imported_reg = self._import_added_endpoint(ed)
            # get exception from ImportRegistration
            exc = imported_reg.get_exception()
            # if there was exception on import, print out messages
            if exc:
                _logger.exception(
                    "BasicTopologyManager import failed for endpoint.id=%s",
                    ed_id,
                )
            else:
                _logger.debug(
                    "BasicTopologyManager: service imported! "
                    "endpoint.id=%s, service_ref=%s",
                    ed_id,
                    imported_reg.get_reference(),
                )
        elif event_type == EndpointEvent.REMOVED:
            self._unimport_removed_endpoint(ed)
            _logger.debug(
                "BasicTopologyManager: endpoint removed. endpoint.id=%s", ed_id
            )
        elif event_type == EndpointEvent.MODIFIED:
            self._update_imported_endpoint(ed)
            _logger.debug(
                "BasicTopologyManager: endpoint updated. endpoint.id=%s", ed_id
            )



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
RequiresBroadcast handler implementation

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import threading

# Pelix beans
from pelix.constants import BundleActivator, BundleException
from pelix.internals.events import ServiceEvent

# iPOPO constants
import pelix.ipopo.constants as ipopo_constants
import pelix.ipopo.handlers.constants as constants
import pelix.ipopo.handlers.requires as requires

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _HandlerFactory(requires._HandlerFactory):
    # pylint: disable=W0212, R0903
    """
    Factory service for service registration handlers
    """

    def get_handlers(self, component_context, instance):
        """
        Sets up service providers for the given component

        :param component_context: The ComponentContext bean
        :param instance: The component instance
        :return: The list of handlers associated to the given component
        """
        # Extract information from the context
        requirements = component_context.get_handler(
            ipopo_constants.HANDLER_REQUIRES_BRODCAST
        )
        requires_filters = component_context.properties.get(
            ipopo_constants.IPOPO_REQUIRES_FILTERS, None
        )

        # Prepare requirements
        requirements = self._prepare_requirements(
            requirements, requires_filters
        )

        # Set up the runtime dependency handlers
        handlers = []
        for field, config in requirements.items():
            # Extract values from tuple
            requirement, muffle_ex, trace_ex = config

            # Construct the handler
            handlers.append(
                BroadcastDependency(field, requirement, muffle_ex, trace_ex)
            )

        return handlers


@BundleActivator
class _Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Sets up members
        """
        self._registration = None

    def start(self, context):
        """
        Bundle started
        """
        # Set up properties
        properties = {
            constants.PROP_HANDLER_ID: ipopo_constants.HANDLER_REQUIRES_BRODCAST
        }

        # Register the handler factory service
        self._registration = context.register_service(
            constants.SERVICE_IPOPO_HANDLER_FACTORY,
            _HandlerFactory(),
            properties,
        )

    def stop(self, _):
        """
        Bundle stopped
        """
        # Unregister the service
        self._registration.unregister()
        self._registration = None


# ------------------------------------------------------------------------------


class _ProxyDummy(object):
    """
    Dummy "Yes Man" object
    """

    def __init__(self, handler, name):
        """
        :param handler: The parent BroadcastHandler
        :param name: Name of this field
        """
        self.__handler = handler
        self.__name = name

    def __bool__(self):
        """
        Returns True if at least one service is bound
        """
        return self.__handler.has_services()

    # Python 2 compatibility
    __nonzero__ = __bool__

    def __call__(self, *args, **kwargs):
        """
        We have to handle a call
        """
        return self.__handler.handle_call(self.__name, args, kwargs)

    def __getattr__(self, member):
        """
        Recursive proxy
        """
        if self.__name is not None:
            member = "{0}.{1}".format(self.__name, member)

        return _ProxyDummy(self.__handler, member)


# ------------------------------------------------------------------------------


class BroadcastDependency(constants.DependencyHandler):
    """
    Manages a required dependency field when a component is running
    """

    def __init__(self, field, requirement, muffle_exceptions, trace_exceptions):
        """
        Sets up the dependency

        :param field: The injected field name
        :param requirement: The Requirement describing this dependency
        :param muffle_exceptions: Flag to not propagate exceptions
        :param proxy_class: Class to use to emulate the missing requirement
        """
        # The internal state lock
        self._lock = threading.RLock()

        # The iPOPO StoredInstance object (given during manipulation)
        self._ipopo_instance = None

        # The bundle context
        self._context = None

        # The associated field
        self._field = field

        # The underlying requirement
        self.requirement = requirement

        # Exception handling flags
        self._muffle_ex = muffle_exceptions
        self._trace_ex = trace_exceptions

        # Injected proxy
        self._proxy = _ProxyDummy(self, None)

        # The logger
        self._logger = logging.getLogger(
            "-".join(("<n/a>", "RequiresBroadcast", field))
        )

        # Reference -> Service
        self._services = {}

        # Length of the future injected list
        self._future_len = 0

    def manipulate(self, stored_instance, component_instance):
        """
        Stores the given StoredInstance bean.

        :param stored_instance: The iPOPO component StoredInstance
        :param component_instance: The component instance
        """
        # Store the stored instance...
        self._ipopo_instance = stored_instance

        # ... and the bundle context
        self._context = stored_instance.bundle_context

        # Reset the logger
        self._logger = logging.getLogger(
            "-".join(
                (self._ipopo_instance.name, "RequiresBroadcast", self._field)
            )
        )

        # Set the default value for the field if it is optional: the proxy
        if self.requirement.optional:
            setattr(component_instance, self._field, self._proxy)

    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        """
        self._services.clear()

        self.requirement = None
        self._services = None
        self._future_len = 0
        self._lock = None
        self._field = None
        self._ipopo_instance = None
        self._context = None
        self._muffle_ex = False
        self._trace_ex = False
        self._proxy = None

    def get_bindings(self):
        """
        Retrieves the list of the references to the bound services

        :return: A list of ServiceReferences objects
        """
        with self._lock:
            return list(self._services.keys())

    def get_field(self):
        """
        Returns the name of the field handled by this handler
        """
        return self._field

    def get_kinds(self):
        """
        Retrieves the kinds of this handler: 'dependency'

        :return: the kinds of this handler
        """
        return (constants.KIND_DEPENDENCY,)

    def get_value(self):
        """
        Retrieves the value to inject in the component

        :return: The value to inject
        """
        if self._future_len > 0 or (
            self.requirement is not None and self.requirement.optional
        ):
            # We got something to work on
            return self._proxy

        # Invalid state
        return None

    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        return (
            self.requirement is not None and self.requirement.optional
        ) or self._future_len > 0

    def has_services(self):
        """
        Indicates if at least one service is bound (used by the proxy)
        """
        return self._future_len > 0

    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            if svc_ref in self._services:
                # We already know this service
                return False

            # Keep track of the service
            svc = self._services[svc_ref] = self._context.get_service(svc_ref)
            self._future_len += 1

            # Bind it
            self._ipopo_instance.bind(self, svc, svc_ref)
            return True

    def on_service_departure(self, svc_ref):
        """
        Called when a service has been unregistered from the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            try:
                svc = self._services[svc_ref]
            except KeyError:
                # Unknown reference
                return None

            # Future length decreases
            self._future_len -= 1

            # Unbind the service first (to keep access during invalidate)
            self._ipopo_instance.unbind(self, svc, svc_ref)

            try:
                # Clean up
                del self._services[svc_ref]
            except KeyError:
                # Ignore, as it might be a side effect
                pass

            return True

    def service_changed(self, event):
        """
        Called by the framework when a service event occurs
        """
        if (
            self._ipopo_instance is None
            or not self._ipopo_instance.check_event(event)
        ):
            # stop() and clean() may have been called after we have been put
            # inside a listener list copy...
            # or we've been told to ignore this event
            return

        # Call sub-methods
        kind = event.get_kind()
        svc_ref = event.get_service_reference()

        if kind == ServiceEvent.REGISTERED:
            # Service coming
            self.on_service_arrival(svc_ref)

        elif kind in (
            ServiceEvent.UNREGISTERING,
            ServiceEvent.MODIFIED_ENDMATCH,
        ):
            # Service gone or not matching anymore
            self.on_service_departure(svc_ref)

    def start(self):
        """
        Starts the dependency manager
        """
        self._context.add_service_listener(
            self, self.requirement.filter, self.requirement.specification
        )

    def stop(self):
        """
        Stops the dependency manager (must be called before clear())

        :return: The removed bindings (list) or None
        """
        self._context.remove_service_listener(self)
        if self._services:
            return [
                (service, reference)
                for reference, service in self._services.items()
            ]

        return None

    def try_binding(self):
        """
        Searches for the required service if needed

        :raise BundleException: Invalid ServiceReference found
        """
        with self._lock:
            if self._services:
                # We already are alive (not our first call)
                # => we are updated through service events
                return

            # Get all matching services
            refs = self._context.get_all_service_references(
                self.requirement.specification, self.requirement.filter
            )
            if not refs:
                # No match found
                return

            results = []
            try:
                # Bind all new reference
                for reference in refs:
                    added = self.on_service_arrival(reference)
                    if added:
                        results.append(reference)
            except BundleException as ex:
                self._logger.debug("Error binding multiple references: %s", ex)

                # Undo what has just been done, ignoring errors
                for reference in results:
                    try:
                        self.on_service_departure(reference)
                    except BundleException as ex2:
                        self._logger.debug("Error cleaning up: %s", ex2)

                del results[:]
                raise

    def handle_call(self, members_str, args, kwargs):
        """
        Handles a call to the proxy
        """
        if members_str:
            all_members = members_str.split(".")
        else:
            all_members = []

        with self._lock:
            if not self._services:
                # Nothing we can do: return False
                return False

            # Copy the list, just in case we have a side effect
            for svc in list(self._services.values()):
                try:
                    # Find the element to call
                    to_call = svc
                    for member in all_members:
                        to_call = getattr(to_call, member)
                except AttributeError:
                    self._logger.warning("%s as no %s member", svc, members_str)
                else:
                    try:
                        # Call it
                        to_call(*args, **kwargs)
                    except Exception as ex:  # pylint:disable=broad-except
                        if not self._muffle_ex:
                            # Propagate if requested
                            raise ex

                        if self._trace_ex:
                            # Log it
                            self._logger.exception(ex)

            # Service have been notified (or failed silently): return True
            return True



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
Report generation service and shell command.

This bundle provides a service and a shell command to generates reports, i.e.
dictionaries containing the description of the current Pelix framework and of
its environement.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import datetime
import inspect
import json
import linecache
import os
import platform
import socket
import sys
import threading
import time

# Pelix
import pelix.constants
from pelix.constants import BundleActivator, BundleException
from pelix.ipopo.constants import use_ipopo
from pelix.shell import SERVICE_SHELL_COMMAND, SERVICE_SHELL_REPORT

# ------------------------------------------------------------------------------

# Public API
__all__ = ("format_frame_info",)

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


def format_frame_info(frame):
    """
    Formats the given stack frame to show its position in the code and
    part of its context

    :param frame: A stack frame
    """
    # Same as in traceback.extract_stack
    line_no = frame.f_lineno
    code = frame.f_code
    filename = code.co_filename
    method_name = code.co_name
    linecache.checkcache(filename)

    try:
        # Try to get the type of the calling object
        instance = frame.f_locals["self"]
        method_name = "{0}::{1}".format(type(instance).__name__, method_name)
    except KeyError:
        # Not called from a bound method
        pass

    # File & line
    output_lines = [
        '  File "{0}", line {1}, in {2}'.format(filename, line_no, method_name)
    ]

    # Arguments
    if frame.f_locals:
        # Pypy keeps f_locals as an empty dictionary
        arg_info = inspect.getargvalues(frame)
        for name in arg_info.args:
            try:
                output_lines.append(
                    "    - {0:s} = {1}".format(name, repr(frame.f_locals[name]))
                )
            except TypeError:
                # Happens in dict/list-comprehensions in Python 2.x
                name = name[0]
                output_lines.append(
                    "    - {0:s} = {1}".format(name, repr(frame.f_locals[name]))
                )

        if arg_info.varargs:
            output_lines.append(
                "    - *{0:s} = {1}".format(
                    arg_info.varargs, frame.f_locals[arg_info.varargs]
                )
            )

        if arg_info.keywords:
            output_lines.append(
                "    - **{0:s} = {1}".format(
                    arg_info.keywords, frame.f_locals[arg_info.keywords]
                )
            )

    # Line block
    lines = _extract_lines(filename, frame.f_globals, line_no, 3)
    if lines:
        output_lines.append("")
        prefix = "      "
        output_lines.append(
            "{0}{1}".format(prefix, "\n{0}".format(prefix).join(lines))
        )
    return "\n".join(output_lines)


def _extract_lines(filename, f_globals, line_no, around):
    """
    Extracts a block of lines from the given file

    :param filename: Name of the source file
    :param f_globals: Globals of the frame of the current code
    :param line_no: Current line of code
    :param around: Number of line to print before and after the current one
    """
    current_line = linecache.getline(filename, line_no, f_globals)
    if not current_line:
        # No data on this line
        return ""

    lines = []
    # Add some lines before
    for pre_line_no in range(line_no - around, line_no):
        pre_line = linecache.getline(filename, pre_line_no, f_globals)
        lines.append("{0}".format(pre_line.rstrip()))

    # The line itself
    lines.append("{0}".format(current_line.rstrip()))

    # Add some lines after
    for pre_line_no in range(line_no + 1, line_no + around + 1):
        pre_line = linecache.getline(filename, pre_line_no, f_globals)
        lines.append("{0}".format(pre_line.rstrip()))

    # Smart left strip
    minimal_tab = None
    for line in lines:
        if line.strip():
            tab = len(line) - len(line.lstrip())
            if minimal_tab is None or tab < minimal_tab:
                minimal_tab = tab

    if minimal_tab > 0:
        lines = [line[minimal_tab:] for line in lines]

    # Add some place for a marker
    marked_line = ">> {0}".format(lines[around])
    lines = ["   {0}".format(line) for line in lines]
    lines[around] = marked_line
    lines.append("")
    return lines


# ------------------------------------------------------------------------------


class _ReportCommands(object):
    """
    Registers report shell commands
    """

    def __init__(self, context):
        """
        Sets up members

        :param context: The bundle context
        """
        self.__context = context

        # Last computed report
        self.__report = None

        # Level -> Methods
        self.__levels = {
            # OS and machine details
            "os": (self.os_details,),
            "os_env": (self.os_env,),
            # Python
            "python": (self.python_details,),
            "python_path": (self.python_path,),
            "python_modules": (self.python_modules,),
            "process": (self.process_details,),
            # Pelix
            "pelix_basic": (self.pelix_infos,),
            "pelix_bundles": (self.pelix_bundles,),
            "pelix_services": (self.pelix_services,),
            # iPOPO
            "ipopo_instances": (self.ipopo_instances,),
            "ipopo_factories": (self.ipopo_factories,),
            # Extra reports
            "threads": (self.threads_list,),
            "network": (self.network_details,),
        }

        # Aliases, to ease the generation of multiple reports at once
        # Alias -> Levels
        self.__aliases = {
            # Full report
            "full": tuple(self.__levels.keys()),
            # Pelix & iPOPO
            "pelix": ("pelix_basic", "pelix_bundles", "pelix_services"),
            "ipopo": ("ipopo_instances", "ipopo_factories"),
            # Application description
            "app": ("os", "process", "python", "python_path", "os_env"),
            # Standard description levels
            "minimal": ("os", "python", "pelix_basic"),
            "standard": (
                "minimal",
                "python_path",
                "python_modules",
                "process",
                "pelix_bundles",
                "ipopo_factories",
            ),
            "debug": ("standard", "pelix_services", "ipopo_instances"),
        }

    @staticmethod
    def get_namespace():
        """
        Retrieves the name space of this command handler
        """
        return "report"

    def get_methods(self):
        """
        Retrieves the list of tuples (command, method) for this command handler
        """
        return [
            ("levels", self.print_levels),
            ("make", self.make_report),
            ("clear", self.clear_report),
            ("show", self.show_report),
            ("write", self.write_report),
        ]

    def get_level_methods(self, level):
        """
        Returns the methods to call for the given level of report

        :param level: The level of report
        :return: The set of methods to call to fill the report
        :raise KeyError: Unknown level or alias
        """
        try:
            # Real name of the level
            return set(self.__levels[level])
        except KeyError:
            # Alias
            result = set()
            for sub_level in self.__aliases[level]:
                result.update(self.get_level_methods(sub_level))
            return result

    def get_levels(self):
        """
        Returns a copy of the dictionary of levels.

        The key is the name of the report level, the value is the tuple of
        methods to call for that level.
        Multiple levels can call the same method.

        :return: A dictionary of lists of methods to call
        """
        return set(self.__levels).union(self.__aliases)

    def print_levels(self, session):
        """
        Lists available levels
        """
        lines = []
        for level in sorted(self.get_levels()):
            methods = sorted(
                method.__name__ for method in self.get_level_methods(level)
            )
            lines.append("- {0}:".format(level))
            lines.append("\t{0}".format(", ".join(methods)))
        session.write_line("\n".join(lines))

    @staticmethod
    def os_details():
        """
        Returns a dictionary containing details about the operating system
        """
        # Compute architecture and linkage
        bits, linkage = platform.architecture()
        results = {
            # Machine details
            "platform.arch.bits": bits,
            "platform.arch.linkage": linkage,
            "platform.machine": platform.machine(),
            "platform.process": platform.processor(),
            "sys.byteorder": sys.byteorder,
            # OS details
            "os.name": os.name,
            "host.name": socket.gethostname(),
            "sys.platform": sys.platform,
            "platform.system": platform.system(),
            "platform.release": platform.release(),
            "platform.version": platform.version(),
            "encoding.filesystem": sys.getfilesystemencoding(),
        }

        # Paths and line separators
        for name in "sep", "altsep", "pathsep", "linesep":
            results["os.{0}".format(name)] = getattr(os, name, None)

        try:
            # Available since Python 3.4
            results["os.cpu_count"] = os.cpu_count()
        except AttributeError:
            results["os.cpu_count"] = None

        try:
            # Only for Unix
            # pylint: disable=E1101
            results["sys.dlopenflags"] = sys.getdlopenflags()
        except AttributeError:
            results["sys.dlopenflags"] = None

        return results

    @staticmethod
    def os_env():
        """
        Returns a copy of the environment variables
        """
        return os.environ.copy()

    @staticmethod
    def process_details():
        """
        Returns details about the current process
        """
        results = {"argv": sys.argv, "working.directory": os.getcwd()}

        # Process ID and execution IDs (UID, GID, Login, ...)
        for key, method in {
            "pid": "getpid",
            "ppid": "getppid",
            "login": "getlogin",
            "uid": "getuid",
            "euid": "geteuid",
            "gid": "getgid",
            "egid": "getegid",
            "groups": "getgroups",
        }.items():
            try:
                results[key] = getattr(os, method)()
            except (AttributeError, OSError):
                results[key] = None
        return results

    @staticmethod
    def network_details():
        """
        Returns details about the network links
        """
        # Get IPv4 details
        ipv4_addresses = [
            info[4][0]
            for info in socket.getaddrinfo(
                socket.gethostname(), None, socket.AF_INET
            )
        ]

        # Add localhost
        ipv4_addresses.extend(
            info[4][0]
            for info in socket.getaddrinfo("localhost", None, socket.AF_INET)
        )

        # Filter addresses
        ipv4_addresses = sorted(set(ipv4_addresses))

        try:
            # Get IPv6 details
            ipv6_addresses = [
                info[4][0]
                for info in socket.getaddrinfo(
                    socket.gethostname(), None, socket.AF_INET6
                )
            ]

            # Add localhost
            ipv6_addresses.extend(
                info[4][0]
                for info in socket.getaddrinfo(
                    "localhost", None, socket.AF_INET6
                )
            )

            # Filter addresses
            ipv6_addresses = sorted(set(ipv6_addresses))
        except (socket.gaierror, AttributeError):
            # AttributeError: AF_INET6 is missing in some versions of Python
            ipv6_addresses = None

        return {
            "IPv4": ipv4_addresses,
            "IPv6": ipv6_addresses,
            "host.name": socket.gethostname(),
            "host.fqdn": socket.getfqdn(),
        }

    @staticmethod
    def python_details():
        """
        Returns a dictionary containing details about the Python interpreter
        """
        build_no, build_date = platform.python_build()
        results = {
            # Version of interpreter
            "build.number": build_no,
            "build.date": build_date,
            "compiler": platform.python_compiler(),
            "branch": platform.python_branch(),
            "revision": platform.python_revision(),
            "implementation": platform.python_implementation(),
            "version": ".".join(str(v) for v in sys.version_info),
            # API version
            "api.version": sys.api_version,
            # Installation details
            "prefix": sys.prefix,
            "base_prefix": getattr(sys, "base_prefix", None),
            "exec_prefix": sys.exec_prefix,
            "base_exec_prefix": getattr(sys, "base_exec_prefix", None),
            # Execution details
            "executable": sys.executable,
            "encoding.default": sys.getdefaultencoding(),
            # Other details, ...
            "recursion_limit": sys.getrecursionlimit(),
        }

        # Threads implementation details
        thread_info = getattr(sys, "thread_info", (None, None, None))
        results["thread_info.name"] = thread_info[0]
        results["thread_info.lock"] = thread_info[1]
        results["thread_info.version"] = thread_info[2]

        # ABI flags (POSIX only)
        results["abiflags"] = getattr(sys, "abiflags", None)

        # -X options (CPython only)
        results["x_options"] = getattr(sys, "_xoptions", None)
        return results

    @staticmethod
    def python_path():
        """
        Returns the content of sys.path
        """
        return {
            "sys.path": sys.path[:],
            "sys.path_hooks": getattr(sys, "path_hooks", None),
            "sys.meta_path": sys.meta_path,
        }

    @staticmethod
    def python_modules():
        """
        Returns the list of Python modules and their file
        """
        imported = {}
        results = {"builtins": sys.builtin_module_names, "imported": imported}
        for module_name, module_ in sys.modules.items():
            if module_name not in sys.builtin_module_names:
                try:
                    imported[module_name] = inspect.getfile(module_)
                except TypeError:
                    imported[
                        module_name
                    ] = "<no file information :: {0}>".format(repr(module_))

        return results

    def pelix_infos(self):
        """
        Basic information about the Pelix framework instance
        """
        framework = self.__context.get_framework()
        return {
            "version": framework.get_version(),
            "properties": framework.get_properties(),
        }

    def pelix_bundles(self):
        """
        List of installed bundles
        """
        framework = self.__context.get_framework()
        return {
            bundle.get_bundle_id(): {
                "name": bundle.get_symbolic_name(),
                "version": bundle.get_version(),
                "state": bundle.get_state(),
                "location": bundle.get_location(),
            }
            for bundle in framework.get_bundles()
        }

    def pelix_services(self):
        """
        List of registered services
        """
        return {
            svc_ref.get_property(pelix.constants.SERVICE_ID): {
                "specifications": svc_ref.get_property(
                    pelix.constants.OBJECTCLASS
                ),
                "ranking": svc_ref.get_property(
                    pelix.constants.SERVICE_RANKING
                ),
                "properties": svc_ref.get_properties(),
                "bundle.id": svc_ref.get_bundle().get_bundle_id(),
                "bundle.name": svc_ref.get_bundle().get_symbolic_name(),
            }
            for svc_ref in self.__context.get_all_service_references(None)
        }

    def ipopo_factories(self):
        """
        List of iPOPO factories
        """
        try:
            with use_ipopo(self.__context) as ipopo:
                return {
                    name: ipopo.get_factory_details(name)
                    for name in ipopo.get_factories()
                }
        except BundleException:
            # iPOPO is not available:
            return None

    def ipopo_instances(self):
        """
        List of iPOPO instances
        """
        try:
            with use_ipopo(self.__context) as ipopo:
                return {
                    instance[0]: ipopo.get_instance_details(instance[0])
                    for instance in ipopo.get_instances()
                }
        except BundleException:
            # iPOPO is not available:
            return None

    @staticmethod
    def threads_list():
        """
        Lists the active threads and their current code line
        """
        results = {}

        # pylint: disable=W0212
        try:
            # Extract frames
            frames = sys._current_frames()

            # Get the thread ID -> Thread mapping
            names = threading._active.copy()
        except AttributeError:
            # Extraction not available
            return results

        # Sort by thread ID
        thread_ids = sorted(frames.keys())
        for thread_id in thread_ids:
            # Get the corresponding stack
            stack = frames[thread_id]

            # Try to get the thread name
            try:
                name = names[thread_id].name
            except KeyError:
                name = "<unknown>"

            trace_lines = []
            frame = stack
            while frame is not None:
                # Store the line information
                trace_lines.append(format_frame_info(frame))

                # Previous frame...
                frame = frame.f_back

            # Construct the thread description
            results[thread_id] = {
                "name": name,
                "stacktrace": "\n".join(reversed(trace_lines)),
            }

        return results

    def make_report(self, session, *levels):
        """
        Prepares the report at the requested level(s)
        """
        if not levels:
            levels = ["full"]

        try:
            # List the methods to call, avoiding double-calls
            methods = set()
            for level in levels:
                methods.update(self.get_level_methods(level))
        except KeyError as ex:
            # Unknown level
            session.write_line("Unknown report level: {0}", ex)
            self.__report = None
        else:
            # Call each method
            self.__report = {method.__name__: method() for method in methods}
            # Describe the report
            self.__report["report"] = {
                "report.levels": levels,
                "time.stamp": time.time(),
                "time.local": str(datetime.datetime.now()),
                "time.utc": str(datetime.datetime.utcnow()),
            }

        return self.__report

    def clear_report(self, _):
        """
        Deletes the report in memory
        """
        self.__report = None

    @staticmethod
    def json_converter(obj):
        """
        Returns the representation string (repr()) for objects that can't be
        converted to JSON
        """
        return str(obj)

    def to_json(self, data):
        """
        Converts the given object to a pretty-formatted JSON string

        :param data: the object to convert to JSON
        :return: A pretty-formatted JSON string
        """
        # Don't forget the empty line at the end of the file
        return (
            json.dumps(
                data,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
                default=self.json_converter,
            )
            + "\n"
        )

    def show_report(self, session, *levels):
        """
        Shows the report that has been generated
        """
        if levels:
            self.make_report(session, *levels)

        if self.__report:
            session.write_line(self.to_json(self.__report))
        else:
            session.write_line("No report to show")

    def write_report(self, session, filename):
        """
        Writes the report in JSON format to the given file
        """
        if not self.__report:
            session.write_line("No report to write down")
            return

        try:
            with open(filename, "w+") as out_file:
                out_file.write(self.to_json(self.__report))
        except IOError as ex:
            session.write_line("Error writing to file: {0}", ex)


# ------------------------------------------------------------------------------


@BundleActivator
class _Activator(object):
    """
    Activator class for Pelix
    """

    def __init__(self):
        """
        Sets up the activator
        """
        self._svc_reg = None

    def start(self, context):
        """
        Bundle starting
        """
        # Prepare the shell utility service
        self._svc_reg = context.register_service(
            [SERVICE_SHELL_COMMAND, SERVICE_SHELL_REPORT],
            _ReportCommands(context),
            {},
        )

    def stop(self, _):
        """
        Bundle stopping
        """
        # Unregister the services
        if self._svc_reg is not None:
            self._svc_reg.unregister()
            self._svc_reg = None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Service providing handler

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging

# Pelix beans
from pelix.constants import BundleActivator, BundleException

# iPOPO constants
import pelix.ipopo.constants as ipopo_constants
import pelix.ipopo.handlers.constants as constants

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _HandlerFactory(constants.HandlerFactory):
    # pylint: disable=R0903
    """
    Factory service for service registration handlers
    """

    def get_handlers(self, component_context, instance):
        """
        Sets up service providers for the given component

        :param component_context: The ComponentContext bean
        :param instance: The component instance
        :return: The list of handlers associated to the given component
        """
        # Retrieve the handler configuration
        provides = component_context.get_handler(
            ipopo_constants.HANDLER_PROVIDES
        )
        if not provides:
            # Nothing to do
            return ()

        # 1 handler per provided service
        return [
            ServiceRegistrationHandler(
                specs, controller, is_factory, is_prototype
            )
            for specs, controller, is_factory, is_prototype in provides
        ]


@BundleActivator
class _Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Sets up members
        """
        self._registration = None

    def start(self, context):
        """
        Bundle started
        """
        # Set up properties
        properties = {
            constants.PROP_HANDLER_ID: ipopo_constants.HANDLER_PROVIDES
        }

        # Register the handler factory service
        self._registration = context.register_service(
            constants.SERVICE_IPOPO_HANDLER_FACTORY,
            _HandlerFactory(),
            properties,
        )

    def stop(self, _):
        """
        Bundle stopped
        """
        # Unregister the service
        self._registration.unregister()
        self._registration = None


# ------------------------------------------------------------------------------


class ServiceRegistrationHandler(constants.ServiceProviderHandler):
    """
    Handles the registration of a service provided by a component
    """

    def __init__(
        self, specifications, controller_name, is_factory, is_prototype
    ):
        """
        Sets up the handler

        :param specifications: The service specifications
        :param controller_name: Name of the associated service controller
                                (can be None)
        :param is_factory: If True, this is a service factory
        :param is_prototype: If True, this is a prototype service factory
        """
        self.specifications = specifications
        self.__controller = controller_name
        self._ipopo_instance = None

        # Controller is "on" by default
        self.__controller_on = True
        self.__validated = False

        # Service factory flag
        self.__is_factory = is_factory
        self.__is_prototype = is_prototype

        # The ServiceRegistration and ServiceReference objects
        self._registration = None
        self._svc_reference = None

    def _field_controller_generator(self):
        """
        Generates the methods called by the injected controller
        """
        # Local variable, to avoid messing with "self"
        stored_instance = self._ipopo_instance

        def get_value(self, name):
            # pylint: disable=W0613
            """
            Retrieves the controller value, from the iPOPO dictionaries

            :param name: The property name
            :return: The property value
            """
            return stored_instance.get_controller_state(name)

        def set_value(self, name, new_value):
            # pylint: disable=W0613
            """
            Sets the property value and trigger an update event

            :param name: The property name
            :param new_value: The new property value
            """
            # Get the previous value
            old_value = stored_instance.get_controller_state(name)
            if new_value != old_value:
                # Update the controller state
                stored_instance.set_controller_state(name, new_value)

            return new_value

        return get_value, set_value

    def manipulate(self, stored_instance, component_instance):
        """
        Manipulates the component instance

        :param stored_instance: The iPOPO component StoredInstance
        :param component_instance: The component instance
        """
        # Store the stored instance
        self._ipopo_instance = stored_instance

        if self.__controller is None:
            # No controller: do nothing
            return

        # Get the current value of the member (True by default)
        controller_value = getattr(component_instance, self.__controller, True)

        # Store the controller value
        stored_instance.set_controller_state(
            self.__controller, controller_value
        )

        # Prepare the methods names
        getter_name = "{0}{1}".format(
            ipopo_constants.IPOPO_CONTROLLER_PREFIX,
            ipopo_constants.IPOPO_GETTER_SUFFIX,
        )
        setter_name = "{0}{1}".format(
            ipopo_constants.IPOPO_CONTROLLER_PREFIX,
            ipopo_constants.IPOPO_SETTER_SUFFIX,
        )

        # Inject the getter and setter at the instance level
        getter, setter = self._field_controller_generator()
        setattr(component_instance, getter_name, getter)
        setattr(component_instance, setter_name, setter)

    def check_event(self, event):
        """
        Tests if the given service event corresponds to the registered service

        :param event: A service event
        :return: True if the given event references the provided service
        """
        return self._svc_reference is not event.get_service_reference()

    def get_kinds(self):
        """
        Retrieves the kinds of this handler: 'service_provider'

        :return: the kinds of this handler
        """
        return (constants.KIND_SERVICE_PROVIDER,)

    def get_service_reference(self):
        """
        Retrieves the reference of the provided service

        :return: A ServiceReference object
        """
        return self._svc_reference

    def on_controller_change(self, name, value):
        """
        Called by the instance manager when a controller value has been
        modified

        :param name: The name of the controller
        :param value: The new value of the controller
        """
        if self.__controller != name:
            # Nothing to do
            return

        # Update the controller value
        self.__controller_on = value
        if value:
            # Controller switched to "ON"
            self._register_service()
        else:
            # Controller switched to "OFF"
            self._unregister_service()

    def on_property_change(self, name, old_value, new_value):
        """
        Called by the instance manager when a component property is modified

        :param name: The changed property name
        :param old_value: The previous property value
        :param new_value: The new property value
        """
        if self._registration is not None:
            # use the registration to trigger the service event
            self._registration.set_properties({name: new_value})

    def post_validate(self):
        """
        Called by the instance manager once the component has been validated
        """
        # Update the validation flag
        self.__validated = True
        self._register_service()

    def pre_invalidate(self):
        """
        Called by the instance manager before the component is invalidated
        """
        # Update the validation flag
        self.__validated = False

        # Force service unregistration
        self._unregister_service()

    def _register_service(self):
        """
        Registers the provided service, if possible
        """
        if (
            self._registration is None
            and self.specifications
            and self.__validated
            and self.__controller_on
        ):
            # Use a copy of component properties
            properties = self._ipopo_instance.context.properties.copy()
            bundle_context = self._ipopo_instance.bundle_context

            # Register the service
            self._registration = bundle_context.register_service(
                self.specifications,
                self._ipopo_instance.instance,
                properties,
                factory=self.__is_factory,
                prototype=self.__is_prototype,
            )
            self._svc_reference = self._registration.get_reference()

            # Notify the component
            self._ipopo_instance.safe_callback(
                ipopo_constants.IPOPO_CALLBACK_POST_REGISTRATION,
                self._svc_reference,
            )

    def _unregister_service(self):
        """
        Unregisters the provided service, if needed
        """
        if self._registration is not None:
            # Ignore error
            try:
                self._registration.unregister()
            except BundleException as ex:
                # Only log the error at this level
                logger = logging.getLogger(
                    "-".join((self._ipopo_instance.name, "ServiceRegistration"))
                )
                logger.error("Error unregistering a service: %s", ex)

            # Notify the component (even in case of error)
            self._ipopo_instance.safe_callback(
                ipopo_constants.IPOPO_CALLBACK_POST_UNREGISTRATION,
                self._svc_reference,
            )

            self._registration = None
            self._svc_reference = None



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
Definition of classes used by the Pelix shell service and its consumers

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2
:status: Alpha

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import sys
import threading

# Pelix
from pelix.utilities import to_bytes, to_str

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

# Before Python 3, input() was raw_input()
if sys.version_info[0] < 3:
    # pylint: disable=E0602,C0103
    safe_input = raw_input
else:
    # pylint: disable=C0103
    safe_input = input

RESULT_VAR_NAME = "?"
""" Name of the result variable """

# ------------------------------------------------------------------------------


class ShellSession(object):
    """
    Represents a shell session. This is the kind of object given as parameter
    to shell commands
    """

    def __init__(self, io_handler, initial_vars=None):
        # type: (IOHandler, dict) -> None
        """
        Sets up the shell session

        :param io_handler:  The I/O handler associated to the session
        :param initial_vars: Initial variables
        """
        # Store parameters
        self._io_handler = io_handler

        if not isinstance(initial_vars, dict):
            initial_vars = {}
        self.__variables = initial_vars.copy()

        # Special variable for the last result
        self.__variables[RESULT_VAR_NAME] = None

        # Set I/O handler methods aliases
        self.write_line = io_handler.write_line
        self.write_line_no_feed = io_handler.write_line_no_feed

        # Those are defined in IOHandler.__init__
        self.write = io_handler.write
        self.flush = io_handler.flush
        self.prompt = io_handler.prompt

    @property
    def variables(self):
        # type: () -> dict
        """
        A copy of the session variables
        """
        return self.__variables.copy()

    @property
    def last_result(self):
        # type: () -> object
        """
        Returns the content of $result
        """
        return self.__variables[RESULT_VAR_NAME]

    def get(self, name):
        # type: (str) -> object
        """
        Returns the value of a variable

        :param name: Name of the variable
        :return: The value of the variable
        :raise KeyError: Unknown name
        """
        return self.__variables[name]

    def set(self, name, value):
        # type: (str, object) -> None
        """
        Sets/overrides the value of a variable

        :param name: Variable name
        :param value: New value
        """
        self.__variables[name] = value

    def unset(self, name):
        # type: (str) -> None
        """
        Unsets the variable with the given name

        :param name: Variable name
        :raise KeyError: Unknown name
        """
        del self.__variables[name]


# ------------------------------------------------------------------------------


class IOHandler(object):
    """
    Handles I/O operations between the command handler and the client
    It automatically converts the given data to bytes in Python 3.
    """

    def __init__(self, in_stream, out_stream, encoding="UTF-8"):
        """
        Sets up the printer

        :param in_stream: Input stream
        :param out_stream: Output stream
        :param encoding: Output encoding
        """
        self.input = in_stream
        self.output = out_stream
        self.encoding = encoding
        try:
            self.out_encoding = self.output.encoding or self.encoding
        except AttributeError:
            self.out_encoding = self.encoding

        # Thread safety
        self.__lock = threading.RLock()

        # Standard behavior
        self.flush = self.output.flush
        self.write = self.output.write

        # Specific behavior
        if sys.version_info[0] >= 3:
            # In Python 3.6, the "mode" field is not available on file-like
            # objects, but the "encoding" field seems to be present only in
            # string compatible ones
            if "b" in getattr(out_stream, "mode", "") or not hasattr(
                out_stream, "encoding"
            ):
                # Bytes conversion
                self.write = self._write_bytes
            else:
                # Strings accepted
                self.write = self._write_str

        # Very specific
        if in_stream is sys.stdin:
            # Warning: conflicts with the console
            self.prompt = safe_input
        else:
            self.prompt = self._prompt

    def _prompt(self, prompt=None):
        """
        Reads a line written by the user

        :param prompt: An optional prompt message
        :return: The read line, after a conversion to str
        """
        if prompt:
            # Print the prompt
            self.write(prompt)
            self.output.flush()

        # Read the line
        return to_str(self.input.readline())

    def _write_bytes(self, data):
        """
        Converts the given data then writes it

        :param data: Data to be written
        :return: The result of ``self.output.write()``
        """
        with self.__lock:
            self.output.write(to_bytes(data, self.encoding))

    def _write_str(self, data):
        """
        Converts the given data then writes it

        :param data: Data to be written
        :return: The result of ``self.output.write()``
        """
        with self.__lock:
            self.output.write(
                to_str(data, self.encoding)
                .encode()
                .decode(self.out_encoding, errors="replace")
            )

    def write_line(self, line=None, *args, **kwargs):
        """
        Formats and writes a line to the output
        """
        if line is None:
            # Empty line
            self.write("\n")
        else:
            # Format the line, if arguments have been given
            if args or kwargs:
                line = line.format(*args, **kwargs)

            with self.__lock:
                # Write it
                self.write(line)
                try:
                    if line[-1] != "\n":
                        # Add the trailing new line
                        self.write("\n")
                except IndexError:
                    # Got an empty string
                    self.write("\n")

        self.flush()

    def write_line_no_feed(self, line=None, *args, **kwargs):
        """
        Formats and writes a line to the output
        """
        if line is None:
            # Empty line
            line = ""
        else:
            # Format the line, if arguments have been given
            if args or kwargs:
                line = line.format(*args, **kwargs)

            # Remove the trailing line feed
            if line[-1] == "\n":
                line = line[:-1]

        # Write it
        self.write(line)
        self.flush()



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
Defines the shell completion handlers for Pelix concepts

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2
:status: Alpha

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import absolute_import

# Try to import readline
try:
    import readline
except ImportError:
    readline = None

# Add some typing
try:
    # pylint: disable=W0611
    from typing import List
    from pelix.framework import BundleContext
    from pelix.shell.beans import ShellSession
except ImportError:
    pass

# Pelix
from pelix.constants import SERVICE_ID, BundleActivator

# Completion classes
from .decorators import SVC_COMPLETER, PROP_COMPLETER_ID, BUNDLE, SERVICE
from .core import Completer

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class BundleCompleter(Completer):
    """
    Completes a bundle ID and display a bundle name in current matches
    """

    @staticmethod
    def display_hook(prompt, session, context, matches, longest_match_len):
        # type: (str, ShellSession, BundleContext, List[str], int) -> None
        """
        Displays the available bundle matches and the bundle name

        :param prompt: Shell prompt string
        :param session: Current shell session (for display)
        :param context: BundleContext of the shell
        :param matches: List of words matching the substitution
        :param longest_match_len: Length of the largest match
        """
        # Prepare a line pattern for each match
        match_pattern = "{{0: >{}}}: {{1}}".format(longest_match_len)

        # Sort matching IDs
        matches = sorted(int(match) for match in matches)

        # Print the match and the associated name
        session.write_line()
        for bnd_id in matches:
            bnd = context.get_bundle(bnd_id)
            session.write_line(match_pattern, bnd_id, bnd.get_symbolic_name())

        # Print the prompt, then current line
        session.write(prompt)
        session.write_line_no_feed(readline.get_line_buffer())
        readline.redisplay()

    def complete(
        self, config, prompt, session, context, current_arguments, current
    ):
        # type: (CompletionInfo, str, ShellSession, BundleContext, List[str], str) -> List[str]
        """
        Returns the list of bundle IDs matching the current state

        :param config: Configuration of the current completion
        :param prompt: Shell prompt (for re-display)
        :param session: Shell session (to display in shell)
        :param context: Bundle context of the Shell bundle
        :param current_arguments: Current arguments (without the command itself)
        :param current: Current word
        :return: A list of matches
        """
        # Register a method to display helpful completion
        self.set_display_hook(self.display_hook, prompt, session, context)

        # Return a list of bundle IDs (strings) matching the current value
        # and not yet in arguments
        rl_matches = []
        for bnd in context.get_bundles():
            bnd_id = "{0} ".format(bnd.get_bundle_id())
            if bnd_id.startswith(current):
                rl_matches.append(bnd_id)

        return rl_matches


class ServiceCompleter(Completer):
    """
    Completes a service ID and display a specification in current matches
    """

    @staticmethod
    def display_hook(prompt, session, context, matches, longest_match_len):
        # type: (str, ShellSession, BundleContext, List[str], int) -> None
        """
        Displays the available services matches and the service details

        :param prompt: Shell prompt string
        :param session: Current shell session (for display)
        :param context: BundleContext of the shell
        :param matches: List of words matching the substitution
        :param longest_match_len: Length of the largest match
        """
        try:
            # Prepare a line pattern for each match
            match_pattern = "{{0: >{}}}: {{1}}".format(longest_match_len)

            # Sort matching IDs
            matches = sorted(int(match) for match in matches)

            # Print the match and the associated name
            session.write_line()
            for svc_id in matches:
                svc_ref = context.get_service_reference(
                    None, "({}={})".format(SERVICE_ID, svc_id)
                )
                session.write_line(match_pattern, svc_id, str(svc_ref))

            # Print the prompt, then current line
            session.write(prompt)
            session.write_line_no_feed(readline.get_line_buffer())
            readline.redisplay()
        except Exception as ex:
            session.write_line("\n{}\n\n", ex)

    def complete(
        self, config, prompt, session, context, current_arguments, current
    ):
        # type: (CompletionInfo, str, ShellSession, BundleContext, List[str], str) -> List[str]
        """
        Returns the list of services IDs matching the current state

        :param config: Configuration of the current completion
        :param prompt: Shell prompt (for re-display)
        :param session: Shell session (to display in shell)
        :param context: Bundle context of the Shell bundle
        :param current_arguments: Current arguments (without the command itself)
        :param current: Current word
        :return: A list of matches
        """
        # Register a method to display helpful completion
        self.set_display_hook(self.display_hook, prompt, session, context)

        # Return a list of bundle IDs (strings) matching the current value
        # and not yet in arguments
        rl_matches = []
        for svc_ref in context.get_all_service_references(None, None):
            svc_id = "{0} ".format(svc_ref.get_property(SERVICE_ID))
            if svc_id.startswith(current):
                rl_matches.append(svc_id)

        return rl_matches


# ------------------------------------------------------------------------------


# All completers for this bundle
COMPLETERS = {BUNDLE: BundleCompleter, SERVICE: ServiceCompleter}


@BundleActivator
class _Activator:
    """
    Bundle activator
    """

    def __init__(self):
        self._registrations = []

    def start(self, context):
        # type: (BundleContext) -> None
        """
        Bundle starting

        :param context: The bundle context
        """
        # Register all completers we know
        self._registrations = [
            context.register_service(
                SVC_COMPLETER,
                completer_class(),
                {PROP_COMPLETER_ID: completer_id},
            )
            for completer_id, completer_class in COMPLETERS.items()
        ]

    def stop(self, _):
        # type: (BundleContext) -> None
        """
        Bundle stopping
        """
        # Clean up
        for svc_reg in self._registrations:
            svc_reg.unregister()
        del self._registrations[:]



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: Specifications handling utility methods

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
try:
    # Python 3
    # pylint: disable=F0401,E0611
    from urllib.parse import urlparse
except ImportError:
    # Python 2
    # pylint: disable=F0401
    from urlparse import urlparse

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Dict, Iterable, List, Optional, Tuple
except ImportError:
    pass

# Pelix
from pelix.utilities import is_string
import pelix.constants
import pelix.framework
import pelix.ldapfilter
import pelix.remote
import pelix.utilities

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

PYTHON_LANGUAGE = "python"
""" Prefix to use for the Python specifications """

# ------------------------------------------------------------------------------


class ExportEndpoint(object):
    """
    Represents an export end point (one per group of configuration types)
    """

    def __init__(
        self, uid, fw_uid, configurations, name, svc_ref, service, properties
    ):
        # type: (str, str, Any[str, List[str]], str, pelix.framework.ServiceReference, object, dict) -> None
        """
        :param uid: Unique identified of the end point
        :param fw_uid: The framework UID
        :param configurations: Kinds of end point (xmlrpc, ...)
        :param name: Name of the end point
        :param svc_ref: ServiceReference of the exported service
        :param service: Instance of the exported service
        :param properties: Extra properties
        :raise ValueError: Invalid UID or the end point exports nothing
                           (all specifications have been filtered)
        """
        if not uid:
            raise ValueError("Invalid UID")

        # Given information
        self.__uid = uid
        self.__fw_uid = fw_uid
        self.__instance = service
        self.__reference = svc_ref
        self.__configurations = configurations
        self.__name = name

        # Normalize extra properties
        if isinstance(properties, dict):
            self.__properties = properties
        else:
            self.__properties = {}

        # Normalize the list of configurations
        if is_string(configurations):
            self.__configurations = (configurations,)
        else:
            self.__configurations = tuple(configurations)

        # Exported specifications
        self.__exported_specs = []  # type: List[str]
        exported_specs = compute_exported_specifications(svc_ref)
        if exported_specs:
            # Transform the specifications for export (add the language prefix)
            self.__exported_specs = format_specifications(exported_specs)
        else:
            raise ValueError(
                "Endpoint {0}, {1}, exports nothing".format(
                    self.__uid, self.__name
                )
            )

    def __hash__(self):
        """
        Custom hash, as we override equality tests
        """
        return hash(self.__uid)

    def __eq__(self, other):
        """
        Equality checked by UID
        """
        return self.__uid == other.uid

    def __ne__(self, other):
        """
        Inequality checked by UID
        """
        return self.__uid != other.uid

    def __str__(self):
        """
        String representation
        """
        return "ExportEndpoint(uid={0}, types={1}, specs={2})".format(
            self.__uid, self.__configurations, self.__exported_specs
        )

    def get_properties(self):
        # type: () -> dict
        """
        Returns merged properties

        :return: Endpoint merged properties
        """
        # Get service properties
        properties = self.__reference.get_properties()

        # Merge with local properties
        properties.update(self.__properties)

        # Some properties can't be merged
        for key in pelix.constants.OBJECTCLASS, pelix.constants.SERVICE_ID:
            properties[key] = self.__reference.get_property(key)

        # Force the exported configurations
        properties[pelix.remote.PROP_EXPORTED_CONFIGS] = self.configurations

        return properties

    def make_import_properties(self):
        # type: () -> dict
        """
        Returns the properties of this endpoint where export properties have
        been replaced by import ones

        :return: A dictionary with import properties
        """
        # Convert merged properties
        props = to_import_properties(self.get_properties())

        # Add the framework UID
        props[pelix.remote.PROP_ENDPOINT_FRAMEWORK_UUID] = self.__fw_uid
        return props

    def rename(self, new_name):
        # type: (str) -> None
        """
        Updates the endpoint name

        :param new_name: The new name of the endpoint
        """
        if new_name:
            # Update the name only if the new one is valid
            self.__name = new_name

    # Access to the service
    @property
    def instance(self):
        # type: () -> object
        """
        Service instance
        """
        return self.__instance

    @property
    def reference(self):
        # type: () -> pelix.framework.ServiceReference
        """
        Service reference
        """
        return self.__reference

    # End point properties
    @property
    def uid(self):
        # type: () -> str
        """
        End point unique identifier
        """
        return self.__uid

    @property
    def framework(self):
        # type: () -> str
        """
        Framework UID
        """
        return self.__fw_uid

    @property
    def configurations(self):
        # type: () ->  Tuple[str, ...]
        """
        Configurations of this end point
        """
        return self.__configurations

    @property
    def name(self):
        # type: () -> str
        """
        Name of the end point
        """
        return self.__name

    @property
    def specifications(self):
        # type: () -> List[str]
        """
        Returns the exported specifications
        """
        return self.__exported_specs


# ------------------------------------------------------------------------------


class ImportEndpoint(object):
    """
    Represents an end point to access an imported service
    """

    def __init__(
        self, uid, framework, configurations, name, specifications, properties
    ):
        # type: (str, str, Any[str, List[str]], Optional[str], List[str], dict) -> None
        """
        :param uid: Unique identified of the end point
        :param framework: UID of the framework exporting the end point
                          (can be None)
        :param configurations: Kinds of end point (xmlrpc, ...)
        :param name: Name of the end point
        :param specifications: Specifications of the exported service
        :param properties: Properties of the service
        """
        self.__uid = uid
        self.__fw_uid = framework or None
        self.__name = name
        self.__properties = properties.copy() if properties else {}

        # Normalize list of configurations
        if is_string(configurations):
            tuple_conf = (configurations,)  # type: Tuple[str, ...]
        else:
            tuple_conf = tuple(configurations)

        self.__configurations = tuple_conf

        # Extract the language prefix in specifications
        self.__specifications = extract_specifications(
            specifications, self.__properties
        )

        # Public variable: the source server,
        # set up by a Pelix discovery service
        self.server = None  # type: str

    def __str__(self):
        """
        String representation of the end point
        """
        return (
            "ImportEndpoint(uid={0}, framework={1}, configurations={2}, "
            "specs={3})".format(
                self.__uid,
                self.__fw_uid,
                self.__configurations,
                self.__specifications,
            )
        )

    # Access to the service details
    @property
    def specifications(self):
        # type: () -> List[str]
        """
        Specifications of the service
        """
        return self.__specifications

    @property
    def properties(self):
        # type: () -> dict
        """
        Properties of the imported service
        """
        return self.__properties

    @properties.setter
    def properties(self, properties):
        # type: (dict) -> None
        """
        Sets the properties of the imported service
        """
        # Keep a copy of the new properties
        self.__properties = properties.copy() if properties else {}

    # End point properties
    @property
    def uid(self):
        # type: () -> str
        """
        End point unique identifier
        """
        return self.__uid

    @property
    def framework(self):
        # type: () -> str
        """
        UID of the framework exporting this end point
        """
        return self.__fw_uid

    @property
    def configurations(self):
        # type: () -> Tuple[str, ...]
        """
        Kind of end point
        """
        return self.__configurations

    @property
    def name(self):
        # type: () -> str
        """
        Name of the end point
        """
        return self.__name


# ------------------------------------------------------------------------------


class EndpointDescription(object):
    """
    Endpoint description bean, according to OSGi specifications:

    http://www.osgi.org/javadoc/r4v42/org/osgi/service/remoteserviceadmin/
    EndpointDescription.html

    This is an importer-side description
    """

    def __init__(self, svc_ref, properties):
        # type: (Optional[pelix.framework.ServiceReference], dict) -> None
        """
        Sets up the description with the given properties

        :raise ValueError: Invalid properties
        """
        # Set up properties
        all_properties = {}  # type: Dict[str, Any]
        if svc_ref is not None:
            all_properties.update(svc_ref.get_properties())

        if properties:
            all_properties.update(properties)

        # Add  some properties if the service reference is given
        if svc_ref is not None:
            # Service ID
            all_properties[
                pelix.remote.PROP_ENDPOINT_SERVICE_ID
            ] = svc_ref.get_property(pelix.constants.SERVICE_ID)

        # Convert properties
        self.__properties = to_import_properties(all_properties)

        # Check their validity
        self.__check_properties(self.__properties)

        # Keep a copy of the endpoint ID
        self.__endpoint_id = self.get_id()

    def __hash__(self):
        """
        Custom hash, as we override equality tests
        """
        return hash(self.__endpoint_id)

    def __eq__(self, other):
        """
        Equality checked by UID
        """
        # pylint: disable=W0212
        return self.__endpoint_id == other.__endpoint_id

    def __ne__(self, other):
        """
        Inequality checked by UID
        """
        # pylint: disable=W0212
        return self.__endpoint_id != other.__endpoint_id

    def __str__(self):
        """
        String representation
        """
        return (
            "EndpointDescription(id={0}; endpoint.service.id={1}; "
            "framework.uuid={2})".format(
                self.get_id(), self.get_service_id(), self.get_framework_uuid()
            )
        )

    @staticmethod
    def __check_properties(props):
        # type: (dict) -> None
        """
        Checks that the given dictionary doesn't have export keys and has
        import keys

        :param props: Properties to validate
        :raise ValueError: Invalid properties
        """
        # Mandatory properties
        mandatory = (
            pelix.remote.PROP_ENDPOINT_ID,
            pelix.remote.PROP_IMPORTED_CONFIGS,
            pelix.constants.OBJECTCLASS,
        )
        for key in mandatory:
            if key not in props:
                raise ValueError("Missing property: {0}".format(key))

        # Export/Import properties
        props_export = (
            pelix.remote.PROP_EXPORTED_CONFIGS,
            pelix.remote.PROP_EXPORTED_INTERFACES,
        )

        for key in props_export:
            if key in props:
                raise ValueError("Export property found: {0}".format(key))

    def get_configuration_types(self):
        # type: () -> List[str]
        """
        Returns the configuration types.

        A distribution provider exports a service with an endpoint.
        This endpoint uses some kind of communications protocol with a set of
        configuration parameters.
        There are many different types but each endpoint is configured by only
        one configuration type.
        However, a distribution provider can be aware of different
        configuration types and provide synonyms to increase the change a
        receiving distribution provider can create a connection to this
        endpoint.
        This value of the configuration types is stored in the
        pelix.remote.PROP_IMPORTED_CONFIGS service property.

        :return: The configuration types (list of str)
        """
        # Return a copy of the list
        return self.__properties[pelix.remote.PROP_IMPORTED_CONFIGS][:]

    def get_framework_uuid(self):
        # type: () -> str
        """
        Returns the UUID of the framework exporting this endpoint, or None

        :return: A framework UUID (str) or None
        """
        return self.__properties.get(pelix.remote.PROP_ENDPOINT_FRAMEWORK_UUID)

    def get_id(self):
        # type: () -> str
        """
        Returns the endpoint's id.
        """
        return self.__properties[pelix.remote.PROP_ENDPOINT_ID]

    def get_intents(self):
        # type: () -> List[str]
        """
        Returns the list of intents implemented by this endpoint.

        The intents are based on the service.intents on an imported service,
        except for any intents that are additionally provided by the importing
        distribution provider.
        All qualified intents must have been expanded.
        This value of the intents is stored in the
        pelix.remote.PROP_INTENTS service property.

        :return: A list of intents (list of str)
        """
        # Return a copy of the list
        try:
            return self.__properties[pelix.remote.PROP_INTENTS][:]
        except KeyError:
            return []

    def get_interfaces(self):
        # type: () -> List[str]
        """
        Provides the list of interfaces implemented by the exported service.

        :return: A list of specifications (list of str)
        """
        return self.__properties[pelix.constants.OBJECTCLASS][:]

    def get_package_version(self, package):
        # type: (str) -> Tuple[int, ...]
        """
        Provides the version of the given package name.

        :param package: The name of the package
        :return: The version of the specified package as a tuple or (0,0,0)
        """
        name = "{0}{1}".format(
            pelix.remote.PROP_ENDPOINT_PACKAGE_VERSION_, package
        )
        try:
            # Get the version string
            version = self.__properties[name]

            # Split dots ('.')
            return tuple(version.split("."))
        except KeyError:
            # No version
            return 0, 0, 0

    def get_properties(self):
        # type: () -> dict
        """
        Returns all endpoint properties.

        :return: A copy of the endpoint properties
        """
        return self.__properties.copy()

    def get_service_id(self):
        # type: () -> int
        """
        Returns the service id for the service exported through this endpoint.

        :return: The ID of service on the exporter side, or 0
        """
        try:
            return self.__properties[pelix.remote.PROP_ENDPOINT_SERVICE_ID]
        except KeyError:
            # Not found
            return 0

    def is_same_service(self, endpoint):
        # type: (EndpointDescription) -> bool
        """
        Tests if this endpoint and the given one have the same framework UUID
        and service ID

        :param endpoint: Another endpoint
        :return: True if both endpoints represent the same remote service
        """
        return (
            self.get_framework_uuid() == endpoint.get_framework_uuid()
            and self.get_service_id() == endpoint.get_service_id()
        )

    def matches(self, ldap_filter):
        # type: (Any[str, pelix.ldapfilter.LDAPFilter]) -> bool
        """
        Tests the properties of this EndpointDescription against the given
        filter

        :param ldap_filter: A filter
        :return: True if properties matches the filter
        """
        return pelix.ldapfilter.get_ldap_filter(ldap_filter).matches(
            self.__properties
        )

    def to_import(self):
        # type: () -> ImportEndpoint
        """
        Converts an EndpointDescription bean to an ImportEndpoint

        :return: An ImportEndpoint bean
        """
        # Properties
        properties = self.get_properties()

        # Framework UUID
        fw_uid = self.get_framework_uuid()

        # Endpoint name
        try:
            # From Pelix UID
            name = properties[pelix.remote.PROP_ENDPOINT_NAME]
        except KeyError:
            # Generated
            name = "{0}.{1}".format(fw_uid, self.get_service_id())

        # Configuration / kind
        configurations = self.get_configuration_types()

        # Interfaces
        specifications = self.get_interfaces()

        return ImportEndpoint(
            self.get_id(),
            fw_uid,
            configurations,
            name,
            specifications,
            properties,
        )

    @classmethod
    def from_export(cls, endpoint):
        # type: (ExportEndpoint) -> EndpointDescription
        """
        Converts an ExportEndpoint bean to an EndpointDescription

        :param endpoint: An ExportEndpoint bean
        :return: An EndpointDescription bean
        """
        assert isinstance(endpoint, ExportEndpoint)

        # Service properties
        properties = endpoint.get_properties()

        # Set import keys
        properties[pelix.remote.PROP_ENDPOINT_ID] = endpoint.uid
        properties[pelix.remote.PROP_IMPORTED_CONFIGS] = endpoint.configurations
        properties[
            pelix.remote.PROP_EXPORTED_INTERFACES
        ] = endpoint.specifications

        # Remove export keys
        for key in (
            pelix.remote.PROP_EXPORTED_CONFIGS,
            pelix.remote.PROP_EXPORTED_INTERFACES,
            pelix.remote.PROP_EXPORTED_INTENTS,
            pelix.remote.PROP_EXPORTED_INTENTS_EXTRA,
        ):
            try:
                del properties[key]
            except KeyError:
                pass

        # Other information
        properties[pelix.remote.PROP_ENDPOINT_NAME] = endpoint.name
        properties[
            pelix.remote.PROP_ENDPOINT_FRAMEWORK_UUID
        ] = endpoint.framework

        return EndpointDescription(None, properties)


# ------------------------------------------------------------------------------


def to_import_properties(properties):
    # type: (dict) -> dict
    """
    Returns a dictionary where export properties have been replaced by import
    ones

    :param properties: A dictionary of service properties (with export keys)
    :return: A dictionary with import properties
    """
    # Copy the given dictionary
    props = properties.copy()

    # Add the "imported" property
    props[pelix.remote.PROP_IMPORTED] = True

    # Remote service ID
    try:
        props[pelix.remote.PROP_ENDPOINT_SERVICE_ID] = props.pop(
            pelix.constants.SERVICE_ID
        )
    except KeyError:
        # No service ID
        pass

    # Replace the "export configs"
    configs = props.pop(pelix.remote.PROP_EXPORTED_CONFIGS, None)
    if configs:
        props[pelix.remote.PROP_IMPORTED_CONFIGS] = configs

    # Clear other export properties
    for key in (
        pelix.remote.PROP_EXPORTED_INTENTS,
        pelix.remote.PROP_EXPORTED_INTENTS_EXTRA,
        pelix.remote.PROP_EXPORTED_INTERFACES,
    ):
        try:
            del props[key]
        except KeyError:
            # Key wasn't there
            pass

    return props


# ------------------------------------------------------------------------------


def compute_exported_specifications(svc_ref):
    # type: (pelix.framework.ServiceReference) -> List[str]
    """
    Computes the list of specifications exported by the given service

    :param svc_ref: A ServiceReference
    :return: The list of exported specifications (or an empty list)
    """
    if svc_ref.get_property(pelix.remote.PROP_EXPORT_NONE):
        # The export of this service is explicitly forbidden, stop here
        return []

    # Service specifications
    specs = svc_ref.get_property(pelix.constants.OBJECTCLASS)

    # Exported specifications
    exported_specs = svc_ref.get_property(pelix.remote.PROP_EXPORTED_INTERFACES)

    if exported_specs and exported_specs != "*":
        # A set of specifications is exported, replace "objectClass"
        iterable_exports = pelix.utilities.to_iterable(exported_specs, False)
        all_exported_specs = [
            spec for spec in specs if spec in iterable_exports
        ]
    else:
        # Export everything
        all_exported_specs = pelix.utilities.to_iterable(specs)

    # Authorized and rejected specifications
    export_only_specs = pelix.utilities.to_iterable(
        svc_ref.get_property(pelix.remote.PROP_EXPORT_ONLY), False
    )

    if export_only_specs:
        # Filter specifications (keep authorized specifications)
        return [
            spec for spec in all_exported_specs if spec in export_only_specs
        ]

    # Filter specifications (reject)
    rejected_specs = pelix.utilities.to_iterable(
        svc_ref.get_property(pelix.remote.PROP_EXPORT_REJECT), False
    )
    return [spec for spec in all_exported_specs if spec not in rejected_specs]


def extract_specifications(specifications, properties):
    # type: (Any[str, List[str]], dict) -> List[str]
    """
    Converts "python:/name" specifications to "name". Keeps the other
    specifications as is.

    :param specifications: The specifications found in a remote registration
    :param properties: Service properties
    :return: The filtered specifications (as a list)
    """
    all_specs = set(pelix.utilities.to_iterable(specifications))
    try:
        synonyms = pelix.utilities.to_iterable(
            properties[pelix.remote.PROP_SYNONYMS], False
        )
        all_specs.update(synonyms)
    except KeyError:
        # No synonyms property
        pass

    filtered_specs = set()
    for original in all_specs:
        try:
            # Extract information
            lang, spec = _extract_specification_parts(original)
            if lang == PYTHON_LANGUAGE:
                # Language match: keep the name only
                filtered_specs.add(spec)
            else:
                # Keep the name as is
                filtered_specs.add(original)
        except ValueError:
            # Ignore invalid specifications
            pass

    return list(filtered_specs)


def format_specifications(specifications):
    # type: (Iterable[str]) -> List[str]
    """
    Transforms the interfaces names into URI strings, with the interface
    implementation language as a scheme.

    :param specifications: Specifications to transform
    :return: The transformed names
    """
    transformed = set()
    for original in specifications:
        try:
            lang, spec = _extract_specification_parts(original)
            transformed.add(_format_specification(lang, spec))
        except ValueError:
            # Ignore invalid specifications
            pass

    return list(transformed)


def _extract_specification_parts(specification):
    # type: (str) -> Tuple[str, str]
    """
    Extract the language and the interface from a "language:/interface"
    interface name

    :param specification: The formatted interface name
    :return: A (language, interface name) tuple
    :raise ValueError: Invalid specification content
    """
    try:
        # Parse the URI-like string
        parsed = urlparse(specification)
    except:
        # Invalid URL
        raise ValueError("Invalid specification URL: {0}".format(specification))

    # Extract the interface name
    interface = parsed.path

    # Extract the language, if given
    language = parsed.scheme
    if not language:
        # Simple name, without scheme
        language = PYTHON_LANGUAGE
    else:
        # Formatted name: un-escape it, without the starting '/'
        interface = _unescape_specification(interface[1:])

    return language, interface


def _format_specification(language, specification):
    # type: (str, str) -> str
    """
    Formats a "language://interface" string

    :param language: Specification language
    :param specification: Specification name
    :return: A formatted string
    """
    return "{0}:/{1}".format(language, _escape_specification(specification))


def _escape_specification(specification):
    # type: (str) -> str
    """
    Escapes the interface string: replaces slashes '/' by '%2F'

    :param specification: Specification name
    :return: The escaped name
    """
    return specification.replace("/", "%2F")


def _unescape_specification(specification):
    # type: (str) -> str
    """
    Unescapes the interface string: replaces '%2F' by slashes '/'

    :param specification: Specification name
    :return: The unescaped name
    """
    return specification.replace("%2F", "/")



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
EventListenerHook for Pelix.

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

try:
    from collections.abc import MutableMapping, MutableSequence
except ImportError:
    from collections import MutableMapping, MutableSequence

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class ShrinkableList(MutableSequence):
    # pylint: disable=R0901
    """
    List where items can be removed, but nothing can be added.
    For use in ShrinkableMap
    """

    def __init__(self, delegate):
        self._delegate = delegate

    def __len__(self):
        return len(self._delegate)

    def __getitem__(self, index):
        return self._delegate[index]

    def __delitem__(self, index):
        del self._delegate[index]

    def __setitem__(self, index, value):
        raise IndexError

    def insert(self, index, value):
        raise IndexError

    def __str__(self):
        return str(self._delegate)


class ShrinkableMap(MutableMapping):
    """
    Map where item->value mappings can be removed, but nothing can be added.
    For use in EventListenerHook
    """

    def __init__(self, delegate):
        self._delegate = delegate

    def __getitem__(self, key):
        return self._delegate[key]

    def __setitem__(self, key, value):
        raise IndexError

    def __delitem__(self, key):
        del self._delegate[key]

    def __iter__(self):
        return self._delegate.__iter__()

    def __len__(self):
        return len(self._delegate)


class ListenerInfo(object):
    """
    Keeps information about a listener
    """

    # Try to reduce memory footprint (stored instances)
    __slots__ = (
        "__bundle_context",
        "__listener",
        "__specification",
        "__ldap_filter",
    )

    def __init__(self, bundle_context, listener, specification, ldap_filter):
        """
        :param bundle_context: Bundle context
        :param listener: Listener instance
        :param specification: Specification to listen to
        :param ldap_filter: LDAP filter on service properties
        """
        self.__bundle_context = bundle_context
        self.__listener = listener
        self.__specification = specification
        self.__ldap_filter = ldap_filter

    @property
    def bundle_context(self):
        """
        The context of the bundle which added the listener.
        """
        return self.__bundle_context

    @property
    def listener(self):
        """
        The listener instance
        """
        return self.__listener

    @property
    def specification(self):
        """
        The specification to listen to
        """
        return self.__specification

    @property
    def ldap_filter(self):
        """
        The LDAP filter on service properties
        """
        return self.__ldap_filter

    def get_bundle_context(self):
        """
        Return the context of the bundle which added the listener.

        :return: A BundleContext object
        """
        return self.__bundle_context

    def get_filter(self):
        """
        Returns the LDAP filter string with which the filter was added

        :return: An LDAP filter string
        """
        if self.__ldap_filter:
            return str(self.__ldap_filter)
        return None


class EventListenerHook(object):
    """
    Event listener hook interface prototype.  The method in this class must be
    overridden for a service event listener hook to be called via whiteboard
    pattern
    """

    def event(self, service_event, listener_dict):
        """
        Method called when a service event is triggered.

        :param service_event: The ServiceEvent being triggered
        :param listener_dict: A dictionary associating a bundle context to a
                              list of listeners
        """
        pass



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Properties handler

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Pelix beans
from pelix.constants import BundleActivator

# iPOPO constants
import pelix.ipopo.constants as ipopo_constants
import pelix.ipopo.handlers.constants as constants

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _HandlerFactory(constants.HandlerFactory):
    # pylint: disable=R0903
    """
    Factory service for service registration handlers
    """

    def get_handlers(self, component_context, instance):
        """
        Sets up service providers for the given component

        :param component_context: The ComponentContext bean
        :param instance: The component instance
        :return: The list/tuple of handlers associated to the given component
        """
        # 1 handler per provided service
        return (PropertiesHandler(),)


@BundleActivator
class _Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Sets up members
        """
        self._registration = None

    def start(self, context):
        """
        Bundle started
        """
        # Set up properties
        properties = {
            constants.PROP_HANDLER_ID: ipopo_constants.HANDLER_PROPERTY
        }

        # Register the handler factory service
        self._registration = context.register_service(
            constants.SERVICE_IPOPO_HANDLER_FACTORY,
            _HandlerFactory(),
            properties,
        )

    def stop(self, _):
        """
        Bundle stopped
        """
        # Unregister the service
        self._registration.unregister()
        self._registration = None


# ------------------------------------------------------------------------------


class PropertiesHandler(constants.Handler):
    """
    Handles the properties
    """

    def __init__(self):
        """
        Sets up the handler
        """
        self._ipopo_instance = None

    def _field_property_generator(self, public_properties):
        """
        Generates the methods called by the injected class properties

        :param public_properties: If True, create a public property accessor,
                                  else an hidden property accessor
        :return: getter and setter methods
        """
        # Local variable, to avoid messing with "self"
        stored_instance = self._ipopo_instance

        # Choose public or hidden properties
        # and select the method to call to notify about the property update
        if public_properties:
            properties = stored_instance.context.properties
            update_notifier = stored_instance.update_property
        else:
            # Copy Hidden properties and remove them from the context
            properties = stored_instance.context.grab_hidden_properties()
            update_notifier = stored_instance.update_hidden_property

        def get_value(_, name):
            """
            Retrieves the property value, from the iPOPO dictionaries

            :param name: The property name
            :return: The property value
            """
            return properties.get(name)

        def set_value(_, name, new_value):
            """
            Sets the property value and trigger an update event

            :param name: The property name
            :param new_value: The new property value
            """
            assert stored_instance.context is not None

            # Get the previous value
            old_value = properties.get(name)
            if new_value != old_value:
                # Change the property
                properties[name] = new_value

                # New value is different of the old one, trigger an event
                update_notifier(name, old_value, new_value)

            return new_value

        return get_value, set_value

    @staticmethod
    def get_methods_names(public_properties):
        """
        Generates the names of the fields where to inject the getter and setter
        methods

        :param public_properties: If True, returns the names of public property
                                  accessors, else of hidden property ones
        :return: getter and a setter field names
        """
        if public_properties:
            prefix = ipopo_constants.IPOPO_PROPERTY_PREFIX
        else:
            prefix = ipopo_constants.IPOPO_HIDDEN_PROPERTY_PREFIX

        return (
            "{0}{1}".format(prefix, ipopo_constants.IPOPO_GETTER_SUFFIX),
            "{0}{1}".format(prefix, ipopo_constants.IPOPO_SETTER_SUFFIX),
        )

    def manipulate(self, stored_instance, component_instance):
        """
        Manipulates the component instance

        :param stored_instance: The iPOPO component StoredInstance
        :param component_instance: The component instance
        """
        # Store the stored instance
        self._ipopo_instance = stored_instance

        # Public flags to generate (True for public accessors)
        flags_to_generate = set()
        if stored_instance.context.properties:
            flags_to_generate.add(True)

        # (False for hidden ones)
        if stored_instance.context.has_hidden_properties():
            flags_to_generate.add(False)

        # Inject properties getters and setters
        for public_flag in flags_to_generate:
            # Prepare methods
            getter, setter = self._field_property_generator(public_flag)

            # Inject the getter and setter at the instance level
            getter_name, setter_name = self.get_methods_names(public_flag)
            setattr(component_instance, getter_name, getter)
            setattr(component_instance, setter_name, setter)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Constants and exceptions for Pelix.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import inspect

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

ACTIVATOR = "__pelix_bundle_activator__"
"""
Name of the module member that will be used as bundle activator.
It must be an object with the following methods:

* start(BundleContext)
* stop(BundleContext)
"""

ACTIVATOR_LEGACY = "activator"
"""
Deprecated: prefer ACTIVATOR

Name of the module member that will be used as bundle activator.
It must be an object with the following methods:

* start(BundleContext)
* stop(BundleContext)
"""

OBJECTCLASS = "objectClass"
"""
Property containing the list of specifications (strings) provided by a service
"""

SERVICE_ID = "service.id"
"""
Property containing the ID of a service.
This ID is unique in a framework instance.
"""

SERVICE_BUNDLEID = "service.bundleid"
"""
Property containing the ID of the bundle providing the service.
"""

SERVICE_PID = "service.pid"
"""
Property containing the Persistent ID of a service, i.e. a string identifier
that will always be the same for a (kind of) service, even after restarting
the framework.
This is used by the Configuration Admin to bind managed services and
configurations.
"""

SERVICE_RANKING = "service.ranking"
"""
Property that indicates the ranking of a service. It is used to sort the
results of methods like get_service_references()
"""

SERVICE_SCOPE = "service.scope"
"""
Property that indicates the service's scope, one of "singleton", "bundle" or
"prototype".
This allows the framework to detect service factories and prototype service
factories.
"""

FRAMEWORK_UID = "framework.uid"
"""
Framework instance "unique" identifier. Used in Remote Services to identify
a framework from another.
It can be generated or be forced using the framework initialization properties.
This property is constant during the life of a framework instance.
"""

OSGI_FRAMEWORK_UUID = "org.osgi.framework.uuid"
"""
OSGi standard framework uuid property name.   Set in framework init to the value
of FRAMEWORK_UID
"""

# ------------------------------------------------------------------------------

SCOPE_SINGLETON = "singleton"
"""
Default service scope: the service is a singleton, which means that the service
object is shared by all bundles
"""

SCOPE_BUNDLE = "bundle"
"""
Service factory scope: the service factory is called each time a new bundle
requires the service.
Each service can have its own version of the service, but will get the same
version if it calls ``get_service()`` multiple times.
"""

SCOPE_PROTOTYPE = "prototype"
"""
Prototype service factory scope: the factory is called each time the caller
gets the service.
This allows all bundles to have multiples objects for the same service.
"""

# ------------------------------------------------------------------------------


def BundleActivator(clazz):
    # pylint: disable=C0103
    """
    Decorator to declare the bundle activator

    Instantiates the decorated class and stores it as a module member.

    :param clazz: The decorated bundle activator class
    :return: The class itself
    """
    # Add the activator instance to the module
    setattr(inspect.getmodule(clazz), ACTIVATOR, clazz())

    # Return the untouched class
    return clazz


# ------------------------------------------------------------------------------


class BundleException(Exception):
    """
    The base of all framework exceptions
    """

    def __init__(self, content):
        """
        Sets up the exception
        """
        if isinstance(content, Exception):
            Exception.__init__(self, str(content))

        else:
            Exception.__init__(self, content)


class FrameworkException(Exception):
    """
    A framework exception is raised when an error can force the framework to
    stop.
    """

    def __init__(self, message, needs_stop=False):
        """
        Sets up the exception

        :param message: A description of the exception
        :param needs_stop: If True, the framework must be stopped
        """
        Exception.__init__(self, message)
        self.needs_stop = needs_stop



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
IPv6 double stack utility module

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import os
import platform
import socket

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


def ipproto_ipv6():
    """
    Returns the value of socket.IPPROTO_IPV6

    :return: The value of socket.IPPROTO_IPV6
    :raise AttributeError: Python or system doesn't support IPv6
    """
    try:
        # pylint: disable=E1101
        return socket.IPPROTO_IPV6
    except AttributeError:
        if os.name == "nt":
            # Known bug: http://bugs.python.org/issue6926
            return 41
        else:
            # Unknown value
            raise


def set_double_stack(socket_obj, double_stack=True):
    # type: (socket.socket, bool) -> None
    """
    Sets up the IPv6 double stack according to the operating system

    :param socket_obj: A socket object
    :param double_stack: If True, use the double stack, else only support IPv6
    :raise AttributeError: Python or system doesn't support V6
    :raise socket.error: Error setting up the double stack value
    """
    try:
        # Use existing value
        opt_ipv6_only = socket.IPV6_V6ONLY
    except AttributeError:
        # Use "known" value
        if os.name == "nt":
            # Windows: see ws2ipdef.h
            opt_ipv6_only = 27
        elif platform.system() == "Linux":
            # Linux: see linux/in6.h (in recent kernels)
            opt_ipv6_only = 26
        else:
            # Unknown value: do nothing
            raise

    # Setup the socket (can raise a socket.error)
    socket_obj.setsockopt(ipproto_ipv6(), opt_ipv6_only, int(not double_stack))



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix shell service bundle.

Provides the basic command parsing and execution support to make a Pelix shell.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import os
import sys
import threading

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Dict, Tuple
except ImportError:
    pass

# Pelix modules
import pelix.constants as constants
import pelix.framework as pelix

# Shell constants
from pelix.shell import (
    SERVICE_SHELL,
    SERVICE_SHELL_COMMAND,
    SERVICE_SHELL_UTILS,
)
from pelix.shell.report import format_frame_info
import pelix.shell.parser as parser

# Shell completion
from pelix.shell.completion import Completion, BUNDLE, SERVICE

# ------------------------------------------------------------------------------

# Public API
__all__ = ()

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _ShellUtils(object):
    """
    Utility methods for the shell
    """

    @staticmethod
    def bundlestate_to_str(state):
        """
        Converts a bundle state integer to a string
        """
        states = {
            pelix.Bundle.INSTALLED: "INSTALLED",
            pelix.Bundle.ACTIVE: "ACTIVE",
            pelix.Bundle.RESOLVED: "RESOLVED",
            pelix.Bundle.STARTING: "STARTING",
            pelix.Bundle.STOPPING: "STOPPING",
            pelix.Bundle.UNINSTALLED: "UNINSTALLED",
        }

        return states.get(state, "Unknown state ({0})".format(state))

    @staticmethod
    def make_table(headers, lines, prefix=None):
        """
        Generates an ASCII table according to the given headers and lines

        :param headers: List of table headers (N-tuple)
        :param lines: List of table lines (N-tuples)
        :param prefix: Optional prefix for each line
        :return: The ASCII representation of the table
        :raise ValueError: Different number of columns between headers and
                           lines
        """
        # Normalize the prefix
        prefix = str(prefix or "")

        # Maximum lengths
        lengths = [len(title) for title in headers]

        # Store the number of columns (0-based)
        nb_columns = len(lengths) - 1

        # Lines
        str_lines = []
        for idx, line in enumerate(lines):
            # Recompute lengths
            str_line = []
            str_lines.append(str_line)
            column = -1

            try:
                for column, entry in enumerate(line):
                    str_entry = str(entry)
                    str_line.append(str_entry)

                    if len(str_entry) > lengths[column]:
                        lengths[column] = len(str_entry)

            except IndexError:
                # Line too small/big
                raise ValueError(
                    "Different sizes for header and lines "
                    "(line {0})".format(idx + 1)
                )

            except (TypeError, AttributeError):
                # Invalid type of line
                raise ValueError(
                    "Invalid type of line: %s", type(line).__name__
                )

            else:
                if column != nb_columns:
                    # Check if all lines have the same number of columns
                    raise ValueError(
                        "Different sizes for header and lines "
                        "(line {0})".format(idx + 1)
                    )

        # Prepare the head (centered text)
        format_str = "{0}|".format(prefix)
        for column, length in enumerate(lengths):
            format_str += " {%d:^%d} |" % (column, length)

        head_str = format_str.format(*headers)

        # Prepare the separator, according the length of the headers string
        separator = "{0}{1}".format(prefix, "-" * (len(head_str) - len(prefix)))
        idx = head_str.find("|")
        while idx != -1:
            separator = "+".join((separator[:idx], separator[idx + 1 :]))
            idx = head_str.find("|", idx + 1)

        # Prepare the output
        output = [separator, head_str, separator.replace("-", "=")]

        # Compute the lines
        format_str = format_str.replace("^", "<")
        for line in str_lines:
            output.append(format_str.format(*line))
            output.append(separator)

        # Force the last end of line
        output.append("")

        # Join'em
        return "\n".join(output)


# ------------------------------------------------------------------------------


class _ShellService(parser.Shell):
    # pylint: disable=R0904
    """
    Provides the core shell service for Pelix
    """

    def __init__(self, context, utilities):
        # type: (pelix.BundleContext, _ShellUtils) -> None
        """
        Sets up the shell

        :param context: The bundle context
        """
        super(_ShellService, self).__init__(context.get_framework(), __name__)
        self._context = context
        self._utils = utilities

        # Bound services: reference -> service
        self._bound_references = {}  # type: Dict[pelix.ServiceReference, Any]

        # Service reference -> (name space, [commands])
        self._reference_commands = {}  # type: Dict[pelix.ServiceReference, Tuple[str, str]]

        # Last working directory
        self._previous_path = None

        # Register basic commands
        self.register_command(None, "bd", self.bundle_details)
        self.register_command(None, "bl", self.bundles_list)

        self.register_command(None, "sd", self.service_details)
        self.register_command(None, "sl", self.services_list)

        self.register_command(None, "start", self.start)
        self.register_command(None, "stop", self.stop)
        self.register_command(None, "update", self.update)
        self.register_command(None, "install", self.install)
        self.register_command(None, "uninstall", self.uninstall)

        self.register_command(None, "properties", self.properties_list)
        self.register_command(None, "property", self.property_value)

        self.register_command(None, "sysprops", self.environment_list)
        self.register_command(None, "sysprop", self.environment_value)

        self.register_command(None, "threads", self.threads_list)
        self.register_command(None, "thread", self.thread_details)

        self.register_command(None, "loglevel", self.log_level)

        self.register_command(None, "cd", self.change_dir)
        self.register_command(None, "pwd", self.print_dir)

    def bind_handler(self, svc_ref):
        """
        Called if a command service has been found.
        Registers the methods of this service.

        :param svc_ref: A reference to the found service
        :return: True if the commands have been registered
        """
        if svc_ref in self._bound_references:
            # Already bound service
            return False

        # Get the service
        handler = self._context.get_service(svc_ref)

        # Get its name space
        namespace = handler.get_namespace()
        commands = []

        # Register all service methods directly
        for command, method in handler.get_methods():
            self.register_command(namespace, command, method)
            commands.append(command)

        # Store the reference
        self._bound_references[svc_ref] = handler
        self._reference_commands[svc_ref] = (namespace, commands)
        return True

    def unbind_handler(self, svc_ref):
        """
        Called if a command service is gone.
        Unregisters its commands.

        :param svc_ref: A reference to the unbound service
        :return: True if the commands have been unregistered
        """
        if svc_ref not in self._bound_references:
            # Unknown reference
            return False

        # Unregister its commands
        namespace, commands = self._reference_commands[svc_ref]
        for command in commands:
            self.unregister(namespace, command)

        # Release the service
        self._context.unget_service(svc_ref)
        del self._bound_references[svc_ref]
        del self._reference_commands[svc_ref]
        return True

    @staticmethod
    def get_banner():
        """
        Returns the Shell banner
        """
        return "** Pelix Shell prompt **\n"

    def var_set(self, session, **kwargs):
        """
        Sets the given variables or prints the current ones. "set answer=42"
        """
        if not kwargs:
            session.write_line(
                self._utils.make_table(
                    ("Name", "Value"), session.variables.items()
                )
            )
        else:
            for name, value in kwargs.items():
                name = name.strip()
                session.set(name, value)
                session.write_line("{0}={1}", name, value)

    @Completion(BUNDLE)
    def bundle_details(self, io_handler, bundle_id):
        """
        Prints the details of the bundle with the given ID or name
        """
        bundle = None

        try:
            # Convert the given ID into an integer
            bundle_id = int(bundle_id)
        except ValueError:
            # Not an integer, suppose it's a bundle name
            for bundle in self._context.get_bundles():
                if bundle.get_symbolic_name() == bundle_id:
                    break
            else:
                # Bundle not found
                bundle = None
        else:
            # Integer ID: direct access
            try:
                bundle = self._context.get_bundle(bundle_id)
            except constants.BundleException:
                pass

        if bundle is None:
            # No matching bundle
            io_handler.write_line("Unknown bundle ID: {0}", bundle_id)
            return False

        lines = [
            "ID......: {0}".format(bundle.get_bundle_id()),
            "Name....: {0}".format(bundle.get_symbolic_name()),
            "Version.: {0}".format(bundle.get_version()),
            "State...: {0}".format(
                self._utils.bundlestate_to_str(bundle.get_state())
            ),
            "Location: {0}".format(bundle.get_location()),
            "Published services:",
        ]
        try:
            services = bundle.get_registered_services()
            if services:
                for svc_ref in services:
                    lines.append("\t{0}".format(svc_ref))
            else:
                lines.append("\tn/a")
        except constants.BundleException as ex:
            # Bundle in a invalid state
            lines.append("\tError: {0}".format(ex))

        lines.append("Services used by this bundle:")
        try:
            services = bundle.get_services_in_use()
            if services:
                for svc_ref in services:
                    lines.append("\t{0}".format(svc_ref))
            else:
                lines.append("\tn/a")
        except constants.BundleException as ex:
            # Bundle in a invalid state
            lines.append("\tError: {0}".format(ex))

        lines.append("")
        io_handler.write("\n".join(lines))
        return None

    def bundles_list(self, io_handler, name=None):
        """
        Lists the bundles in the framework and their state. Possibility to
        filter on the bundle name.
        """
        # Head of the table
        headers = ("ID", "Name", "State", "Version")

        # Get the bundles
        bundles = self._context.get_bundles()

        # The framework is not in the result of get_bundles()
        bundles.insert(0, self._context.get_framework())

        if name is not None:
            # Filter the list
            bundles = [
                bundle
                for bundle in bundles
                if name in bundle.get_symbolic_name()
            ]

        # Make the entries
        lines = [
            [
                str(entry)
                for entry in (
                    bundle.get_bundle_id(),
                    bundle.get_symbolic_name(),
                    self._utils.bundlestate_to_str(bundle.get_state()),
                    bundle.get_version(),
                )
            ]
            for bundle in bundles
        ]

        # Print'em all
        io_handler.write(self._utils.make_table(headers, lines))

        if name is None:
            io_handler.write_line("{0} bundles installed", len(lines))
        else:
            io_handler.write_line("{0} filtered bundles", len(lines))

    @Completion(SERVICE)
    def service_details(self, io_handler, service_id):
        """
        Prints the details of the service with the given ID
        """
        svc_ref = self._context.get_service_reference(
            None, "({0}={1})".format(constants.SERVICE_ID, service_id)
        )
        if svc_ref is None:
            io_handler.write_line("Service not found: {0}", service_id)
            return False

        lines = [
            "ID............: {0}".format(
                svc_ref.get_property(constants.SERVICE_ID)
            ),
            "Rank..........: {0}".format(
                svc_ref.get_property(constants.SERVICE_RANKING)
            ),
            "Specifications: {0}".format(
                svc_ref.get_property(constants.OBJECTCLASS)
            ),
            "Bundle........: {0}".format(svc_ref.get_bundle()),
            "Properties....:",
        ]
        for key, value in sorted(svc_ref.get_properties().items()):
            lines.append("\t{0} = {1}".format(key, value))

        lines.append("Bundles using this service:")
        for bundle in svc_ref.get_using_bundles():
            lines.append("\t{0}".format(bundle))

        lines.append("")
        io_handler.write("\n".join(lines))
        return None

    def services_list(self, io_handler, specification=None):
        """
        Lists the services in the framework. Possibility to filter on an exact
        specification.
        """
        # Head of the table
        headers = ("ID", "Specifications", "Bundle", "Ranking")

        # Lines
        references = (
            self._context.get_all_service_references(specification, None) or []
        )

        # Construct the list of services
        lines = [
            [
                str(entry)
                for entry in (
                    ref.get_property(constants.SERVICE_ID),
                    ref.get_property(constants.OBJECTCLASS),
                    ref.get_bundle(),
                    ref.get_property(constants.SERVICE_RANKING),
                )
            ]
            for ref in references
        ]

        if not lines and specification:
            # No matching service found
            io_handler.write_line("No service provides '{0}'", specification)
            return False

        # Print'em all
        io_handler.write(self._utils.make_table(headers, lines))
        io_handler.write_line("{0} services registered", len(lines))
        return None

    def properties_list(self, io_handler):
        """
        Lists the properties of the framework
        """
        # Get the framework
        framework = self._context.get_framework()

        # Head of the table
        headers = ("Property Name", "Value")

        # Lines
        lines = [item for item in framework.get_properties().items()]

        # Sort lines
        lines.sort()

        # Print the table
        io_handler.write(self._utils.make_table(headers, lines))

    def property_value(self, io_handler, name):
        """
        Prints the value of the given property, looking into
        framework properties then environment variables.
        """
        value = self._context.get_property(name)
        if value is None:
            # Avoid printing "None"
            value = ""

        io_handler.write_line(str(value))

    def environment_list(self, io_handler):
        """
        Lists the framework process environment variables
        """
        # Head of the table
        headers = ("Environment Variable", "Value")

        # Lines
        lines = [item for item in os.environ.items()]

        # Sort lines
        lines.sort()

        # Print the table
        io_handler.write(self._utils.make_table(headers, lines))

    @staticmethod
    def environment_value(io_handler, name):
        """
        Prints the value of the given environment variable
        """
        io_handler.write_line(os.getenv(name))

    @staticmethod
    def threads_list(io_handler, max_depth=1):
        """
        Lists the active threads and their current code line
        """
        # Normalize maximum depth
        try:
            max_depth = int(max_depth)
            if max_depth < 1:
                max_depth = None
        except (ValueError, TypeError):
            max_depth = None

        # pylint: disable=W0212
        try:
            # Extract frames
            frames = sys._current_frames()

            # Get the thread ID -> Thread mapping
            names = threading._active.copy()
        except AttributeError:
            io_handler.write_line("sys._current_frames() is not available.")
            return

        # Sort by thread ID
        thread_ids = sorted(frames.keys())
        lines = []
        for thread_id in thread_ids:
            # Get the corresponding stack
            stack = frames[thread_id]

            # Try to get the thread name
            try:
                name = names[thread_id].name
            except KeyError:
                name = "<unknown>"

            # Construct the code position
            lines.append("Thread ID: {0} - Name: {1}".format(thread_id, name))
            lines.append("Stack Trace:")

            trace_lines = []
            depth = 0
            frame = stack
            while frame is not None and (
                max_depth is None or depth < max_depth
            ):
                # Store the line information
                trace_lines.append(format_frame_info(frame))

                # Previous frame...
                frame = frame.f_back
                depth += 1

            # Reverse the lines
            trace_lines.reverse()

            # Add them to the printed lines
            lines.extend(trace_lines)
            lines.append("")

        lines.append("")

        # Sort the lines
        io_handler.write("\n".join(lines))

    @staticmethod
    def thread_details(io_handler, thread_id, max_depth=0):
        """
        Prints details about the thread with the given ID (not its name)
        """
        # Normalize maximum depth
        try:
            max_depth = int(max_depth)
            if max_depth < 1:
                max_depth = None
        except (ValueError, TypeError):
            max_depth = None

        # pylint: disable=W0212
        try:
            # Get the stack
            thread_id = int(thread_id)
            stack = sys._current_frames()[thread_id]
        except KeyError:
            io_handler.write_line("Unknown thread ID: {0}", thread_id)
        except ValueError:
            io_handler.write_line("Invalid thread ID: {0}", thread_id)
        except AttributeError:
            io_handler.write_line("sys._current_frames() is not available.")
        else:
            # Get the name
            try:
                name = threading._active[thread_id].name
            except KeyError:
                name = "<unknown>"

            lines = [
                "Thread ID: {0} - Name: {1}".format(thread_id, name),
                "Stack trace:",
            ]

            trace_lines = []
            depth = 0
            frame = stack
            while frame is not None and (
                max_depth is None or depth < max_depth
            ):
                # Store the line information
                trace_lines.append(format_frame_info(frame))

                # Previous frame...
                frame = frame.f_back
                depth += 1

            # Reverse the lines
            trace_lines.reverse()

            # Add them to the printed lines
            lines.extend(trace_lines)

            lines.append("")
            io_handler.write("\n".join(lines))

    @staticmethod
    def log_level(io_handler, level=None, name=None):
        """
        Prints/Changes log level
        """
        # Get the logger
        logger = logging.getLogger(name)

        # Normalize the name
        if not name:
            name = "Root"

        if not level:
            # Level not given: print the logger level
            io_handler.write_line(
                "{0} log level: {1} (real: {2})",
                name,
                logging.getLevelName(logger.getEffectiveLevel()),
                logging.getLevelName(logger.level),
            )
        else:
            # Set the logger level
            try:
                logger.setLevel(level.upper())
                io_handler.write_line("New level for {0}: {1}", name, level)
            except ValueError:
                io_handler.write_line("Invalid log level: {0}", level)

    def change_dir(self, session, path):
        """
        Changes the working directory
        """
        if path == "-":
            # Previous directory
            path = self._previous_path or "."

        try:
            previous = os.getcwd()
            os.chdir(path)
        except IOError as ex:
            # Can't change directory
            session.write_line("Error changing directory: {0}", ex)
        else:
            # Store previous path
            self._previous_path = previous
            session.write_line(os.getcwd())

    @staticmethod
    def print_dir(session):
        """
        Prints the current working directory
        """
        pwd = os.getcwd()
        session.write_line(pwd)
        return pwd

    def __get_bundle(self, io_handler, bundle_id):
        """
        Retrieves the Bundle object with the given bundle ID. Writes errors
        through the I/O handler if any.

        :param io_handler: I/O Handler
        :param bundle_id: String or integer bundle ID
        :return: The Bundle object matching the given ID, None if not found
        """
        try:
            bundle_id = int(bundle_id)
            return self._context.get_bundle(bundle_id)
        except (TypeError, ValueError):
            io_handler.write_line("Invalid bundle ID: {0}", bundle_id)
        except constants.BundleException:
            io_handler.write_line("Unknown bundle: {0}", bundle_id)

    @Completion(BUNDLE, multiple=True)
    def start(self, io_handler, bundle_id, *bundles_ids):
        """
        Starts the bundles with the given IDs. Stops on first failure.
        """
        for bid in (bundle_id,) + bundles_ids:
            try:
                # Got an int => it's a bundle ID
                bid = int(bid)
            except ValueError:
                # Got something else, we will try to install it first
                bid = self.install(io_handler, bid)

            bundle = self.__get_bundle(io_handler, bid)
            if bundle is not None:
                io_handler.write_line(
                    "Starting bundle {0} ({1})...",
                    bid,
                    bundle.get_symbolic_name(),
                )
                bundle.start()
            else:
                return False

        return None

    @Completion(BUNDLE, multiple=True)
    def stop(self, io_handler, bundle_id, *bundles_ids):
        """
        Stops the bundles with the given IDs. Stops on first failure.
        """
        for bid in (bundle_id,) + bundles_ids:
            bundle = self.__get_bundle(io_handler, bid)
            if bundle is not None:
                io_handler.write_line(
                    "Stopping bundle {0} ({1})...",
                    bid,
                    bundle.get_symbolic_name(),
                )
                bundle.stop()
            else:
                return False

        return None

    @Completion(BUNDLE, multiple=True)
    def update(self, io_handler, bundle_id, *bundles_ids):
        """
        Updates the bundles with the given IDs. Stops on first failure.
        """
        for bid in (bundle_id,) + bundles_ids:
            bundle = self.__get_bundle(io_handler, bid)
            if bundle is not None:
                io_handler.write_line(
                    "Updating bundle {0} ({1})...",
                    bid,
                    bundle.get_symbolic_name(),
                )
                bundle.update()
            else:
                return False

        return None

    def install(self, io_handler, module_name):
        """
        Installs the bundle with the given module name
        """
        bundle = self._context.install_bundle(module_name)
        io_handler.write_line("Bundle ID: {0}", bundle.get_bundle_id())
        return bundle.get_bundle_id()

    @Completion(BUNDLE, multiple=True)
    def uninstall(self, io_handler, bundle_id, *bundles_ids):
        """
        Uninstalls the bundles with the given IDs. Stops on first failure.
        """
        for bid in (bundle_id,) + bundles_ids:
            bundle = self.__get_bundle(io_handler, bid)
            if bundle is not None:
                io_handler.write_line(
                    "Uninstalling bundle {0} ({1})...",
                    bid,
                    bundle.get_symbolic_name(),
                )
                bundle.uninstall()
            else:
                return False

        return None


# ------------------------------------------------------------------------------


@constants.BundleActivator
class _Activator(object):
    """
    Activator class for Pelix
    """

    def __init__(self):
        """
        Sets up the activator
        """
        self._shell = None
        self._shell_reg = None
        self._utils_reg = None
        self._logger = logging.getLogger(__name__)

    def service_changed(self, event):
        # type: (pelix.ServiceEvent) -> None
        """
        Called when a command provider service event occurred
        """
        kind = event.get_kind()
        reference = event.get_service_reference()

        if kind in (pelix.ServiceEvent.REGISTERED, pelix.ServiceEvent.MODIFIED):
            # New or modified service
            self._shell.bind_handler(reference)
        else:
            # Service gone or not matching anymore
            self._shell.unbind_handler(reference)

    def start(self, context):
        # type: (pelix.BundleContext) -> None
        """
        Bundle starting

        :param context: The bundle context
        """
        try:
            # Prepare the shell utility service
            utils = _ShellUtils()
            self._shell = _ShellService(context, utils)
            self._shell_reg = context.register_service(
                SERVICE_SHELL, self._shell, {}
            )
            self._utils_reg = context.register_service(
                SERVICE_SHELL_UTILS, utils, {}
            )

            # Register the service listener
            context.add_service_listener(self, None, SERVICE_SHELL_COMMAND)

            # Register existing command services
            refs = context.get_all_service_references(SERVICE_SHELL_COMMAND)
            if refs is not None:
                for ref in refs:
                    self._shell.bind_handler(ref)

            self._logger.info("Shell services registered")

        except constants.BundleException as ex:
            self._logger.exception(
                "Error registering the shell service: %s", ex
            )

    def stop(self, context):
        # type: (pelix.BundleContext) -> None
        """
        Bundle stopping

        :param context: The bundle context
        """
        # Unregister the service listener
        context.remove_service_listener(self)

        # Unregister the services
        if self._shell_reg is not None:
            self._shell_reg.unregister()
            self._shell_reg = None

        if self._utils_reg is not None:
            self._utils_reg.unregister()
            self._utils_reg = None

        self._shell = None
        self._logger.info("Shell services unregistered")



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""

Topology Manager API

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging

try:
    from typing import Any, Dict
except ImportError:
    pass

from pelix.framework import ServiceEvent, ServiceReference, BundleContext
from pelix.internals.hooks import EventListenerHook
from pelix.ipopo.decorators import Validate, Invalidate, Provides, Requires
from pelix.services import SERVICE_EVENT_LISTENER_HOOK

from pelix.rsa.providers.discovery import (
    SERVICE_ENDPOINT_ADVERTISER,
    SERVICE_ENDPOINT_LISTENER,
    EndpointEventListener,
    EndpointEvent,
)

from pelix.rsa import (
    SERVICE_EXPORTED_INTERFACES,
    get_exported_interfaces,
    SERVICE_RSA_EVENT_LISTENER,
    SERVICE_REMOTE_SERVICE_ADMIN,
    ECF_ENDPOINT_CONTAINERID_NAMESPACE,
    RemoteServiceAdminListener,
    RemoteServiceAdminEvent,
    ImportRegistration,
)
from pelix.rsa.endpointdescription import EndpointDescription

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Standard logging
_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


@Provides(
    [
        SERVICE_EVENT_LISTENER_HOOK,
        SERVICE_RSA_EVENT_LISTENER,
        SERVICE_ENDPOINT_LISTENER,
    ]
)
@Requires("_rsa", SERVICE_REMOTE_SERVICE_ADMIN)
@Requires("_advertisers", SERVICE_ENDPOINT_ADVERTISER, True, True)
class TopologyManager(
    EventListenerHook, RemoteServiceAdminListener, EndpointEventListener, object
):
    def __init__(self):
        self._advertisers = []
        self._context = None  # type: BundleContext
        self._rsa = None

    @Validate
    def _validate(self, context):
        # type: (BundleContext) -> None
        self._context = context

    @Invalidate
    def _invalidate(self, _):
        self._context = None

    def _import_added_endpoint(self, endpoint_description):
        # type: (EndpointDescription) -> ImportRegistration
        return self._rsa.import_service(endpoint_description)

    def _unimport_removed_endpoint(self, endpoint_description):
        # type: (EndpointDescription) -> None
        # pylint: disable=W0212
        import_regs = self._rsa._get_import_regs()
        for import_reg in import_regs:
            if import_reg.match_ed(endpoint_description):
                import_reg.close()

    def _update_imported_endpoint(self, endpoint_description):
        # type: (EndpointDescription) -> None
        # pylint: disable=W0212
        import_regs = self._rsa._get_import_regs()
        for import_reg in import_regs:
            if import_reg.match_ed(endpoint_description):
                import_reg.update(endpoint_description)

    def _handle_service_registered(self, service_ref):
        # type: (ServiceReference) -> None
        exp_intfs = get_exported_interfaces(service_ref)
        # If no exported interfaces, then all done
        if not exp_intfs:
            return
        self._rsa.export_service(
            service_ref, {SERVICE_EXPORTED_INTERFACES: exp_intfs}
        )

    def _handle_service_unregistering(self, service_ref):
        # type: (ServiceReference) -> None
        # pylint: disable=W0212
        export_regs = self._rsa._get_export_regs()
        if export_regs:
            for export_reg in export_regs:
                if export_reg.match_sr(service_ref):
                    _logger.debug(
                        "handle_service_unregistering. closing "
                        "export_registration for service reference=%s",
                        service_ref,
                    )
                    export_reg.close()

    def _handle_service_modified(self, service_ref):
        # type: (ServiceReference) -> EndpointDescription
        # pylint: disable=W0212
        export_regs = self._rsa._get_export_regs()
        if export_regs:
            for export_reg in export_regs:
                if export_reg.match_sr(service_ref):
                    _logger.debug(
                        "_handle_service_modified. updating "
                        "export_registration for service reference=%s",
                        service_ref,
                    )

                    # actually update the export_reg here
                    if not export_reg.update(None):
                        _logger.warning(
                            "_handle_service_modified. updating"
                            "update for service_ref=%s failed",
                            service_ref,
                        )

    def _handle_event(self, service_event):
        # type: (ServiceEvent) -> None
        kind = service_event.get_kind()
        service_ref = service_event.get_service_reference()
        if kind == ServiceEvent.REGISTERED:
            self._handle_service_registered(service_ref)
        elif kind == ServiceEvent.UNREGISTERING:
            self._handle_service_unregistering(service_ref)
        elif kind == ServiceEvent.MODIFIED:
            self._handle_service_modified(service_ref)

    # impl of EventListenerHook
    def event(self, service_event, listener_dict):
        # type: (ServiceEvent, Dict[Any, Any]) -> None
        self._handle_event(service_event)

    def _advertise_endpoint(self, ed):
        # type: (EndpointDescription) -> None
        for adv in self._advertisers:
            try:
                adv.advertise_endpoint(ed)
            except:
                _logger.exception(
                    "Exception in advertise_endpoint for "
                    "advertiser=%s endpoint=%s",
                    adv,
                    ed,
                )

    def _update_endpoint(self, ed):
        # type: (EndpointDescription) -> None
        for adv in self._advertisers:
            try:
                adv.update_endpoint(ed)
            except:
                _logger.exception(
                    "Exception in update_endpoint for advertiser=%s "
                    "endpoint=%s",
                    adv,
                    ed,
                )

    def _unadvertise_endpoint(self, ed):
        # type: (EndpointDescription) -> None
        for adv in self._advertisers:
            try:
                adv.unadvertise_endpoint(ed.get_id())
            except:
                _logger.exception(
                    "Exception in unadvertise_endpoint for advertiser=%s "
                    "endpoint=%s",
                    adv,
                    ed,
                )

    # impl of RemoteServiceAdminListener
    def remote_admin_event(self, rsa_event):
        # type: (RemoteServiceAdminEvent) -> None
        kind = rsa_event.get_type()
        if kind == RemoteServiceAdminEvent.EXPORT_REGISTRATION:
            self._advertise_endpoint(rsa_event.get_description())
        elif kind == RemoteServiceAdminEvent.EXPORT_UNREGISTRATION:
            self._unadvertise_endpoint(rsa_event.get_description())
        elif kind == RemoteServiceAdminEvent.EXPORT_UPDATE:
            self._update_endpoint(rsa_event.get_description())

    def endpoint_changed(self, endpoint_event, matched_filter):
        # type: (EndpointEvent, Any) -> None
        _logger.debug(
            "TopologyManager.endpoint_event called. "
            "You probably want to override this method"
        )
